/*****************************************************************************/
/*!
 * \file theory_core.cpp
 * 
 * Author: Clark Barrett, Vijay Ganesh (CNF converter)
 * 
 * Created: Thu Jan 30 16:57:52 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/


#include "theory_core.h"
#include "vcl.h"
#include "notifylist.h"
#include "search.h"
#include "common_proof_rules.h"
#include "exception.h"
#include "parser_exception.h"
#include "typecheck_exception.h"
#include "smtlib_exception.h"
#include "eval_exception.h"
#include "debug.h"
#include "command_line_flags.h"
#include "pretty_printer.h"
#include "expr_stream.h"



using namespace std;
namespace CVCL {

  //! Implementation of PrettyPrinter class
  /*! \ingroup PrettyPrinting */
  class PrettyPrinterCore: public PrettyPrinter {
  private:
    TheoryCore *d_core;
    //! Disable the default constructor
    PrettyPrinterCore() { }
  public:
    //! Constructor
    PrettyPrinterCore(TheoryCore* core): d_core(core) { }
    ExprStream& print(ExprStream& os, const Expr& e) {
      if(e.isString())
	return e.print(os);
      else if(d_core->hasTheory(e.getKind()))
	return d_core->theoryOf(e.getKind())->print(os, e);
      else
	return e.print(os);
    }
  };

}

using namespace CVCL;

void
TheoryCore::enqueueSE(const Theorem& thm) {
  DebugAssert(d_inAddFact || d_inCheckSATCore, "enqueueSE()");
  if(thm.isRewrite() || !thm.getExpr().isTrue())
    d_queueSE.push_back(thm);
}


Theorem TheoryCore::getModelValue(const Expr& e) {
  ExprHashMap<Theorem>::iterator i=d_varAssignments.find(e),
    iend=d_varAssignments.end();
  if(i!=iend) return (*i).second;
  else return find(e);
}

///////////////////////////////////////////////////////////////////////////////
/*!
 * Function: TheoryCore::TheoryCore
 *
 * Author: Clark Barrett
 * 
 * Created: Sat Feb  8 14:57:36 2003
 *
 * Description: Constructor.  Since we are in the middle of creating
 * TheoryCore, we set the pointer to TheoryCore in the Theory base
 * class ourselves.
*/
///////////////////////////////////////////////////////////////////////////////
TheoryCore::TheoryCore(VCL *vcl)
  : Theory(), d_inconsistent(vcl->getCurrentContext(), false, 0), 
    d_incomplete(vcl->getCurrentContext()),
    d_diseq(vcl->getCurrentContext()),
    d_incThm(vcl->getCurrentContext()),
    d_terms(vcl->getCurrentContext()),
    d_vars(vcl->getCurrentContext()),
    d_sharedTerms(vcl->getCurrentContext()),
    d_solver(NULL),
    d_simplifyInPlace(&(vcl->getFlags()["ip"].getBool())),
    d_cnfOption(&(vcl->getFlags()["cnf"].getBool())),
    d_resourceLimit(0),
#ifdef DEBUG
    d_dumpTrace(&(vcl->getFlags()["dump-trace"].getString())),
    d_inCheckSATCore(false), d_inAddFact(false), d_inSimplify(false),
#endif
    d_notifyObj(this, vcl->getCurrentContext())
{
  IF_DEBUG(d_diseq.setName("CDList[TheoryCore::d_diseq]"));
  d_theoryCore = this;
  d_commonRules = newCommonRules(vcl->getTM());
  d_em = vcl->getEM();
  d_printer = new PrettyPrinterCore(this);
  d_vcl = vcl;
  d_name = "Core";

  // Register the pretty-printer
  d_em->registerPrettyPrinter(*d_printer);

  // for (int i = 0; i < LAST_KIND; ++i) d_theoryMap[i] = NULL;

  vector<int> kinds;
  kinds.push_back(RAW_LIST);
  kinds.push_back(BOOLEAN);
  kinds.push_back(SUBTYPE);
  kinds.push_back(STRING_EXPR);
  kinds.push_back(ID);
  kinds.push_back(TRUE);
  kinds.push_back(FALSE);
  kinds.push_back(UCONST);
  kinds.push_back(BOUND_VAR);
  kinds.push_back(SKOLEM_VAR);
  kinds.push_back(EQ);
  kinds.push_back(NEQ);
  kinds.push_back(ECHO);
  kinds.push_back(DBG);
  kinds.push_back(TRACE);
  kinds.push_back(UNTRACE);
  kinds.push_back(OPTION);
  kinds.push_back(HELP);
  kinds.push_back(AND);
  kinds.push_back(OR);
  kinds.push_back(IFTHEN);
  kinds.push_back(IF);
  kinds.push_back(ELSE);
  kinds.push_back(COND);
  kinds.push_back(XOR);
  kinds.push_back(NOT);
  kinds.push_back(ITE);
  kinds.push_back(IFF);
  kinds.push_back(IMPLIES);
  kinds.push_back(BOOL_VAR);
  // For printing LET expressions (in DAG printing mode)
  kinds.push_back(LET);
  kinds.push_back(LETDECLS);
  kinds.push_back(LETDECL);
  // For printing raw parsed quantifier expressions
  kinds.push_back(VARLIST);
  kinds.push_back(VARDECLS);
  kinds.push_back(VARDECL);

  // Type declarations and definitions
  kinds.push_back(TYPE);
  // For printing type declarations (or definitions)
  kinds.push_back(CONST);

  kinds.push_back(CONSTDEF);
  kinds.push_back(TYPEDEF);
  kinds.push_back(DEFUN);
  // Printing proofs
  kinds.push_back(PF_APPLY);
  kinds.push_back(PF_HOLE);
  // Register commands for pretty-printing.  Currently, only ASSERT
  // needs to be printed.
  kinds.push_back(ASSERT);
  kinds.push_back(QUERY);
  kinds.push_back(PRINT);

  kinds.push_back(DUMP_PROOF);
  kinds.push_back(DUMP_ASSUMPTIONS);
  kinds.push_back(DUMP_SIG);
  kinds.push_back(DUMP_TCC);
  kinds.push_back(DUMP_TCC_ASSUMPTIONS);
  kinds.push_back(DUMP_TCC_PROOF);
  kinds.push_back(DUMP_CLOSURE);
  kinds.push_back(DUMP_CLOSURE_PROOF);
  kinds.push_back(TRANSFORM);
  kinds.push_back(CALL);
  kinds.push_back(WHERE);
  kinds.push_back(COUNTEREXAMPLE);
  kinds.push_back(PUSH);
  kinds.push_back(POP);
  kinds.push_back(POPTO);
  kinds.push_back(PUSH_SCOPE);
  kinds.push_back(POP_SCOPE);
  kinds.push_back(POPTO_SCOPE);
  kinds.push_back(CONTEXT);
  kinds.push_back(FORGET);
  kinds.push_back(GET_TYPE);
  kinds.push_back(CHECK_TYPE);
  kinds.push_back(GET_CHILD);
  kinds.push_back(SUBSTITUTE);

  
  kinds.push_back(AND_R);
  kinds.push_back(IFF_R);
  kinds.push_back(ITE_R);

  registerKinds(this, kinds);
}


TheoryCore::~TheoryCore()
{
  delete d_commonRules;
  d_em->unregisterPrettyPrinter();
  delete d_printer;
}


///////////////////////////////////////////////////////////////////////////////
// Theory interface implementaion                                            //
///////////////////////////////////////////////////////////////////////////////

void TheoryCore::assertFact(const Theorem& e) {
  const Expr& expr = e.getExpr();
  switch(expr.getKind()) {
  case BOOL_VAR: break; // Ignore it
  case NOT:
    DebugAssert(expr[0].getKind()==BOOL_VAR, "TheoryCore::assertFact("
		+expr.toString()+")");
    break; // Ignore that too
  default:
    DebugAssert(false, "TheoryCore::assertFact("
		+expr.toString()+")");
  }
}


void TheoryCore::checkSat(bool fullEffort) {
  // Handle all asserted disequalities
  for (CDList<Theorem>::const_iterator i = d_diseq.begin(), iend=d_diseq.end();
       i!=iend; ++i) {
    const Expr& expr = i->getExpr()[0];
    Theorem thm0 = find(expr[0]);
    Theorem thm1 = find(expr[1]);
    if (thm0.getRHS() == thm1.getRHS()) {
      Theorem leftEqRight = transitivityRule(thm0, symmetryRule(thm1));
      IF_DEBUG(debugger.counter("conflicts from disequalities")++);
      setInconsistent(iffMP(leftEqRight, notToIff(*i)));
      return;
    }
  }
}


Theorem TheoryCore::rewriteLitCore(const Expr& e)
{
  switch (e.getKind()) {
    case EQ:
      if (e[0] == e[1])
        return d_commonRules->rewriteReflexivity(e);
      else if (e[0] < e[1])
        return d_commonRules->rewriteUsingSymmetry(e);
      break;
    case NOT:
      if (e[0].isTrue())
        return d_commonRules->rewriteNotTrue(e);
      else if (e[0].isFalse())
        return d_commonRules->rewriteNotFalse(e);
      else if (e[0].isNot())
        return d_commonRules->rewriteNotNot(e);
      break;
    default:
      DebugAssert(false,
		  "TheoryCore::rewriteLitCore("
		  + e.toString()
		  + "): Not implemented");
      break;
  }
  return reflexivityRule(e);
}


Theorem TheoryCore::rewrite(const Expr& e) {
  Theorem thm;
  switch (e.getKind()) {
    case TRUE:
    case FALSE:
    case UCONST:
    case BOUND_VAR:
      thm = reflexivityRule(e);
      break; // do not rewrite
      // Replace LETDECL and CONSTDEF with their definitions.  The
      // typechecker makes sure it's type-safe to do so.
    case LETDECL:
      thm = d_commonRules->rewriteLetDecl(e);
      break;
    case CONSTDEF:
      thm = d_commonRules->rewriteConstDef(e);
      break;
    case EQ:
    case NOT:
      thm = rewriteLitCore(e);
      break;
    case IMPLIES: {
      thm = d_commonRules->rewriteImplies(e);
      const Expr& rhs = thm.getRHS();
      // rhs = OR(!e1, e2).  Rewrite !e1, then top-level OR().
      DebugAssert(rhs.isOr() && rhs.arity() == 2,
		  "TheoryCore::rewrite[IMPLIES]: rhs = "+rhs.toString());
      Theorem rw = rewriteCore(rhs[0]);
      if(rw.getLHS() != rw.getRHS()) {
	vector<unsigned> changed;
	vector<Theorem> thms;
	changed.push_back(0);
	thms.push_back(rw);
	rw = substitutivityRule(rhs, changed, thms);
	// Simplify to the find pointer of the result
	rw = transitivityRule(rw, find(rw.getRHS()));
	// Now rw = Theorem(rhs = rhs')
	rw = transitivityRule(rw, rewrite(rw.getRHS()));
      } else
	rw = rewrite(rhs);
      thm = transitivityRule(thm, rw);
//       thm = transitivityRule(thm, simplifyThm(thm.getRHS()));
      break;
    }
    case IFF: {
      thm = d_commonRules->rewriteIff(e);
      const Expr& rhs = thm.getRHS();
      // The only time we need to rewrite the result (rhs) is when
      // e==(FALSE<=>e1) or (e1<=>FALSE), so rhs==!e1.
      if (e != rhs && rhs.isNot())
	thm = transitivityRule(thm, rewrite(rhs));
      break;
    }
    case ITE:
      if (e[0].isTrue())
        thm = d_commonRules->rewriteIteTrue(e);
      else if (e[0].isFalse())
        thm = d_commonRules->rewriteIteFalse(e);
      else if (e[1] == e[2])
        thm = d_commonRules->rewriteIteSame(e);
      else if (d_vcl->getFlags()["un-ite-ify"].getBool())
      {
	// undo the rewriting of Boolean connectives to ITEs.
	// helpful for examples converted from SVC.
	// must rewrite again because we might create expressions
	// that can be further rewritten, and we must normalize.
	if (e[1].isFalse() && e[2].isTrue())
	  thm = rewriteCore(d_commonRules->rewriteIteToNot(e));
	else if (e[1].isTrue())
	  thm = rewriteCore(d_commonRules->rewriteIteToOr(e));
	else if (e[2].isFalse())
	  thm = rewriteCore(d_commonRules->rewriteIteToAnd(e));
	else if (e[2].isTrue())
	  thm = rewriteCore(d_commonRules->rewriteIteToImp(e));
	else if (e[1] == e[2].negate())
	  thm = rewriteCore(d_commonRules->rewriteIteToIff(e));
	else thm = reflexivityRule(e);
      }
      else thm = reflexivityRule(e);

      break;
    case AND:
      thm = d_commonRules->rewriteAnd(e);
      break;
    case OR:
      thm = d_commonRules->rewriteOr(e);
      break;
    // Quantifiers
    case FORALL:
    case EXISTS:
    // Boolean variables
    case BOOL_VAR:
      return d_commonRules->reflexivityRule(e);

    // don't need to rewrite these
    case AND_R:
    case IFF_R:
    case ITE_R:
      thm = reflexivityRule(e);
      break;

    default:
      DebugAssert(false,
		  "TheoryCore::rewrite("
		  + e.toString() + " : " + getType(e).toString()
		  + "): Not implemented");
      break;
  }

  const Expr& rhs = thm.getRHS();
  if (theoryOf(rhs) == this) {
    // Core rewrites are idempotent
    rhs.setRewriteNormal();
  }

  return thm;
}


Theorem TheoryCore::simplifyOp(const Expr& e) {
  int kind(e.getKind());
  Theorem res;
  switch(kind) {
  case AND:
  case OR: {
    // Stop when a child has this kind
    int endKind = (kind==AND)? FALSE : TRUE;
    int ar = e.arity();
    // Optimization: before simplifying anything recursively, check if
    // any kid is already TRUE or FALSE, and just return at that point
    int l(0);
    for(; l<ar && e[l].getKind() != endKind; ++l);
    if(l < ar) { // Found TRUE or FALSE: e simplifies to a const
      IF_DEBUG(debugger.counter("simplified AND/OR topdown")++);
      if(kind==AND)
	res = rewriteAnd(e);
      else
	res = rewriteOr(e);
      break; // Break out of switch()
    }
    vector<Theorem> newChildrenThm;
    vector<unsigned> changed;
    for(int k = 0; k < ar; ++k) {
      // Recursively simplify the kids
      Theorem thm = simplify(e[k]);
      if (thm.getLHS() != thm.getRHS()) {
	if (thm.getRHS().getKind() == endKind) {
	  newChildrenThm.clear();
	  changed.clear();
	  newChildrenThm.push_back(thm);
	  changed.push_back(k);
	  thm = substitutivityRule(e, changed, newChildrenThm);
	  // Simplify to TRUE or FALSE
	  if(kind==AND)
	    thm = transitivityRule(thm, rewriteAnd(thm.getRHS()));
	  else
	    thm = transitivityRule(thm, rewriteOr(thm.getRHS()));
	  IF_DEBUG(debugger.counter("simplified AND/OR: skipped kids")
		   += ar-k-1);
	  return thm;
	} else { // Child simplified to something else
	  newChildrenThm.push_back(thm);
	  changed.push_back(k);
	}
      }
    }
    if(changed.size() > 0)
      res = substitutivityRule(e, changed, newChildrenThm);
    break;
  }
  case ITE: {
    DebugAssert(e.arity()==3, "Bad ITE in TheoryCore::simplify(e="
		+e.toString()+")");
    // Optimization: check if the two branches are the same, so we
    // don't have to simplify the condition
    if(e[1]==e[2]) {
      res = d_commonRules->rewriteIteSame(e);
      res = transitivityRule(res, simplify(res.getRHS()));
      IF_DEBUG(debugger.counter("simplified ITE(c,e,e)")++);
      break; // break out of switch()
    }
    // First, simplify the conditional
    vector<Theorem> newChildrenThm;
    vector<unsigned> changed;
    Theorem thm = simplify(e[0]);
    if(thm.getLHS() != thm.getRHS()) {
      newChildrenThm.push_back(thm);
      changed.push_back(0);
    }
    const Expr& cond = thm.getRHS();
    for(int k=1; k<=2; ++k) {
      // If condition value is known, only the appropriate branch
      // needs to be simplified
      if((k==1 && cond.isFalse()) || (k==2 && cond.isTrue())) {
	IF_DEBUG(debugger.counter("simplified ITE: skiped one branch")++);
	continue;
      }
      thm = simplify(e[k]);
      if(thm.getLHS() != thm.getRHS()) {
	newChildrenThm.push_back(thm);
	changed.push_back(k);
      }
    }
    if(changed.size() > 0)
      res = substitutivityRule(e, changed, newChildrenThm);
    break;
  }
  default:
    res = Theory::simplifyOp(e);
  }
  if(res.isNull())
    res = reflexivityRule(e);
  return res;
}


/*! Only Boolean constants (TRUE and FALSE) and variables of
  uninterpreted types should appear here (they come from records and
  tuples).  Just ignore them. */
void TheoryCore::setup(const Expr& e) { }

Expr TheoryCore::computeTypePred(const Type& t, const Expr& e)
{
  Expr tExpr = t.getExpr();
  switch(tExpr.getKind()) {
  case SUBTYPE: {
    Expr pred = tExpr[0];
    // pred may be a named function (CONSTDEF).  Strip CONSTDEFs off.
    while(pred.getKind()==CONSTDEF) pred = pred[1];
    const Type& argTp = pred.lookupType()[0];
    return newExpr(getOp(pred), e).andExpr(getTypePred(argTp, e));
  }
  default:
    return e.getEM()->trueExpr();
  }
}

void TheoryCore::computeType(const Expr& e)
{
  switch (e.getKind()) {
    case ITE: {
      Type t1(getBaseType(e[1])), t2(getBaseType(e[2]));
      if (getType(e[0]) != boolType())
	throw TypecheckException
	  ("The conditional in IF-THEN-ELSE must be BOOLEAN, but is:\n\n"
	   +getType(e[0]).toString()
	   +"\n\nIn the expression:\n\n  "
	   +e.toString());
      if(t1 != t2) {
	throw TypecheckException
	  ("The types of the IF-THEN-ELSE branches do not match.\n"
	   "THEN branch has the type:\n\n  "
	   +getType(e[1]).toString()
	   +"\n\nELSE branch has the type:\n\n  "
	   +getType(e[2]).toString()
	   +"\n\nIn expression:\n\n  "+e.toString());
      }
      Type res(getType(e[1]));
      // If the precise types match in both branches, use it as the
      // result type.
      if(res == getType(e[2])) {
	IF_DEBUG(if(res != t1) // Trace ITEs with strict subtypes
		 TRACE("ite", "ITE: setting type "+res.toString()
		       +" to:\n\n  ", e, ""));
	e.setType(res);
      }
      else
	// Note: setting the base type, since e[1] and e[2] have
	// different exact types, and the base type is a conservative
	// approximation we can easily compute.
	e.setType(t1);
    }
      break;
    case EQ: {
      Type t0(getBaseType(e[0])), t1(getBaseType(e[1]));
      if (t0.isBool() || t1.isBool()) {
	throw TypecheckException
	  ("Cannot use EQ ('=') for BOOLEAN type; use IFF ('<=>') instead.\n"
	   "Error in the following expression:\n"+e.toString());
      }
      if (t0 != t1) {
	throw TypecheckException
	  ("Type mismatch in equality:\n\n LHS type:\n"+t0.toString()
	   +"\n\n RHS type: \n"+t1.toString()
	   +"\n\n in expression: \n"+e.toString());
      }
      e.setType(boolType());
    }
      break;
    case NOT:
    case AND:
    case OR:
    case XOR:
    case IFF:
    case IMPLIES:

    case AND_R:
    case IFF_R:
    case ITE_R:
    
      for (int k = 0; k < e.arity(); ++k) {
	if (getType(e[k]) != boolType()) {
	  throw TypecheckException(e.toString());
	}
      }
      e.setType(boolType());
      break;
    case CONSTDEF: 
    case LETDECL: {
      Type varTp(getBaseType(e[0]));
      Type valTp(getBaseType(e[1]));
      if(valTp != varTp) {
	throw TypecheckException("Type mismatch for "+e[0].toString()+":"
				 +"\n  declared: "
				 + varTp.toString()
				 +"\n  derived: "+ valTp.toString());
      }
      e.setType(getType(e[0]));
    }
      break;
    default:
      DebugAssert(false,"TheoryCore::computeType(" + e.toString()
		  + "):\nNot implemented");
      break;
  }
}


Type
TheoryCore::computeBaseType(const Type& tp) {
  const Expr& e = tp.getExpr();
  Type res;
  switch(e.getKind()) {
  case SUBTYPE: {
    DebugAssert(e.arity() == 1, "Expr::computeBaseType(): "+e.toString());
    Type lambdaTp = getType(e[0]);
    Type lambdaBaseTp = getBaseType(lambdaTp);
    DebugAssert(lambdaBaseTp.isFunction(), 
		"Expr::computeBaseType(): lambdaBaseTp = "
		+lambdaBaseTp.toString()+" in e = "+e.toString());
    res = lambdaBaseTp[0];
    break;
  }
  case BOOLEAN:
    res = tp; break;
  case TYPEDEF: // Compute the base type of the definition
    res = getBaseType(Type(e[1]));
    break;
  default:
    DebugAssert(false, "TheoryCore::computeBaseType("+tp.toString()+")");
    res = tp;
  }
  return res;
}


Expr 
TheoryCore::computeTCC(const Expr& e) {
  Expr res;
  switch (e.getKind()) {
  case NOT:
    res = getTCC(e[0]);
    break;
  case AND: {
    // ( (tcc(e1) & !e1) \/ ... \/ (tcc(en) & !en) \/ (tcc(e1)&...&tcc(en))
    vector<Expr> tccs;
    for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i)
      tccs.push_back(getTCC(*i));
    vector<Expr> pairs;
    pairs.push_back(rewriteAnd(andExpr(tccs)).getRHS());
    for(size_t i=0, iend=tccs.size(); i<iend; ++i)
      pairs.push_back(rewriteAnd(tccs[i].andExpr(e[i].negate())).getRHS());
    res = rewriteOr(orExpr(pairs)).getRHS();
    break;
  }
  case OR: {
    // ( (tcc(e1) & e1) \/ ... \/ (tcc(en) & en) \/ (tcc(e1)&...&tcc(en))
    vector<Expr> tccs;
    for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i)
      tccs.push_back(getTCC(*i));
    vector<Expr> pairs;
    pairs.push_back(rewriteAnd(andExpr(tccs)).getRHS());
    for(size_t i=0, iend=tccs.size(); i<iend; ++i)
      pairs.push_back(rewriteAnd(tccs[i].andExpr(e[i])).getRHS());
    res = rewriteOr(orExpr(pairs)).getRHS();
    break;
  }
  case ITE: {
    Expr tcc1(getTCC(e[1])), tcc2(getTCC(e[2]));
    // Optimize: if TCCs on both branches are the same, skip the ITE
    Expr tccITE((tcc1 == tcc2)? tcc1 : e[0].iteExpr(tcc1, tcc2));
    res = rewriteAnd(getTCC(e[0]).andExpr(tccITE)).getRHS();
    break;
  }
  case IMPLIES:
    res = getTCC(e[0].negate().orExpr(e[1]));
    break;
  default: // All the other operators are strict
    res = Theory::computeTCC(e);
    break;
  }
  return res;
}

///////////////////////////////////////////////////////////////////////////////
// Pretty-printing			                                     //
///////////////////////////////////////////////////////////////////////////////

ExprStream&
TheoryCore::print(ExprStream& os, const Expr& e)   
{
  switch(os.lang()) {
  case PRESENTATION_LANG:
    switch(e.getKind()) {
    case TYPE:
      if(e.arity() == 0) os << "TYPE";
      else if(e.arity() == 1)
	os << e[0] << ": TYPE;";
      else if(e.arity() == 2)
	os << e[0] << ":" << push << " TYPE = " << e[1] << push << ";";
      else e.print(os);
      break;
    case ID:
      if(e.arity() == 1 && e[0].isString()) os << e[0].getString();
      else e.print(os);
      break;
    case BOOL_VAR:
      if(e.isBoolVar()) os << e.getVar();
      else e.print(os);
      break;
    case CONST:
      if(e.arity() == 2) {
	if(e[0].getKind() == RAW_LIST) {
	  bool first(true);
	  for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i) {
	    if(first) first = false;
	    else os << push << "," << pop << space;
	    os << (*i);
	  }
	}
	else
	  os << e[0];
	os << ":" << push << space << e[1] << push << ";";
      } else if(e.arity() == 3)
	os << e[0] << ":" << push << space << e[1]
	   << space << "=" << space << push << e[2] << push << ";";
      else
	e.print(os);
      break;
    case SUBTYPE:
      os << "SUBTYPE(" << push << e[0] << push << ")";
      break;
    case TYPEDEF: {
      // This is used when dumping declarations to file.  Print just
      // the name of the type, unless it's a messed-up expression.
      if(e.arity() != 2) e.print(os);
      else if(e[0].isString()) os << e[0].getString();
      else e[0].print(os);
      break;
    }
    case EQ:
      // When a separator starts with a space (like " = "), add the
      // leading space with 'space' modifier.  If this separator goes
      // to the next line, the leading spaces must be eaten up to get
      // the indentation right; 'space' will tell the indentation
      // engine that it is a space that can be eaten.  A space in a
      // string (like " ") will never be eaten.
      os << "(" << push << e[0] << space << "= " << e[1] << push << ")";
      break;
    case NOT: os << "NOT " << e[0]; break;
    case AND: {
      int i=0, iend=e.arity();
      os << "(" << push;
      if(i!=iend) { os << e[i]; ++i; }
      for(; i!=iend; ++i) os << space << "AND " << e[i];
      os << push << ")";
    }
      break;
    case OR: {
      int i=0, iend=e.arity();
      os << "(" << push;
      if(i!=iend) { os << e[i]; ++i; }
      for(; i!=iend; ++i) os << space << "OR " << e[i];
      os << push << ")";
    }
      break;
    case ITE:
      os << push << "IF " << push << e[0] << popSave
	 << space << "THEN " << pushRestore << e[1] << popSave
	 << space << "ELSE " << pushRestore << e[2] << pop
	 << space << "ENDIF";
      break;
    case IFF:
      if(e.arity() == 2)
	os << "(" << push << e[0] << space << "<=> " << e[1] << push << ")";
      else
	e.print(os);
      break;
    case IMPLIES:
      os << "(" << push << e[0] << space << "=> " << e[1] << push << ")";
      break;
      // Commands
    case ASSERT:
      os << "ASSERT " << push << e[0] << push << ";";
      break;
    case TRANSFORM:
      os << "TRANSFORM " << push << e[0] << push << ";";
      break;
    case QUERY:
      os << "QUERY " << push << e[0] << push << ";";
      break;
    case COUNTEREXAMPLE:
      os << "COUNTEREXAMPLE;";
      break;
    case PUSH:
      if(e.arity()==0)
	os << "PUSH;";
      else
	os << "PUSH" << space << e[0] << push << ";";
      break;
    case POP:
      if(e.arity()==0)
	os << "POP;";
      else
	os << "POP" << space << e[0] << push << ";";
      break;
    case POPTO:
      os << "POPTO" << space << e[0] << push << ";"; break;
    case PUSH_SCOPE:
      if(e.arity()==0)
	os << "PUSH_SCOPE;";
      else
	os << "PUSH_SCOPE" << space << e[0] << push << ";";
      break;
    case POP_SCOPE:
      if(e.arity()==0)
	os << "POP_SCOPE;";
      else
	os << "POP_SCOPE" << space << e[0] << push << ";";
      break;
    case POPTO_SCOPE:
      os << "POPTO_SCOPE" << space << e[0] << push << ";"; break;
    case CONSTDEF:
    case LETDECL:
      if(e.arity() == 2) os << e[1];
      else e.print(os);
      break;
    case LET: {
      // (LET (LETDECLS (LETDECL var [ type ] val) .... ) body)
      bool first(true);
      os << "(" << push << "LET" << space << push;
      for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i) {
	if(!first) os << push << "," << pop << endl;
	else first = false;
	if(i->arity() == 3) {
	  os << (*i)[0] << ":" << space << push << (*i)[1]
	     << space << "= " << push << nodag << (*i)[2] << pop << pop;
	} else {
	  os << (*i)[0];
	  Type tp((*i)[0].lookupType());
	  if(!tp.isNull()) os << ":" << space << push << tp.getExpr();
	  else os << push;
	  os << space << "= " << push << nodag << (*i)[1] << pop << pop;
	}
      }
      os << pop << endl << "IN" << space << push << e[1] << push << ")";
      break;
    }
    case BOUND_VAR:
      os << e.getName()+"_"+e.getUid();
      break;
    case SKOLEM_VAR:
      os << "SKOLEM_" + int2string(e.getIndex());
      break;
    case PF_APPLY: // FIXME: this will eventually go to the "symsim" theory
      DebugAssert(e.arity() > 0, "TheoryCore::print(): "
		  "Proof rule application must have at "
		  "least one argument (rule name):\n "+e.toString());
      os << e[0];
      if(e.arity() > 1) { // Print the arguments
	os << push << "(" << push;
	bool first(true);
	for(int i=1; i<e.arity(); i++) {
	  if(first) first=false;
	  else os << push << "," << pop << space;
	  os << e[i];
	}
	os << push << ")";
      }
      break;
    case RAW_LIST: {
      os << "[" << push;
      bool firstTime(true);
      for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i) {
	if(firstTime) firstTime = false;
	else os << push << "," << pop << space;
	os << *i;
      }
      os << push << "]";
      break;
    }
    case PF_HOLE: // FIXME: implement this (now fall through to default)
    default:
      // Print the top node in the default LISP format, continue with
      // pretty-printing for children.
      e.print(os);
    }
    break; // end of case PRESENTATION_LANG
  case SMTLIB_LANG:
    switch(e.getKind()) {
    case TRUE: os << "true"; break;
    case FALSE: os << "false"; break;
    case UCONST: e.print(os); break;
    case TYPE:
      if (e.arity() == 1) {
        os << "  :extrasorts (";
        for (int i=0; i < e[0].arity(); ++i) {
          if (i != 0) os << push << space;
          os << push << e[0][i];
        }
        os << push << ")";
      }
      else if (e.arity() == 2) {
        break;
      }
      else {
        throw SmtlibException("TheoryCore::print: SMTLIB: Unexpected TYPE expression");
      }
      break;
    case ID: // FIXME: when lisp becomes case-insensitive, fix printing of IDs
      if(e.arity() == 1 && e[0].isString()) os << e[0].getString();
      else e.print(os);
      break;
    case BOOL_VAR:
      if(e.isBoolVar()) os << e.getVar();
      else e.print(os);
      break;
    case CONST:
      if(e.arity() == 2) {
        if (e[1].getKind() == BOOLEAN) {
          os << "  :extrapreds ((" << push << e[0]
             << push << "))";
        }
        else if (e[1].getKind() == ARROW && e[1][e[1].arity()-1].getKind() == BOOLEAN) {
          os << "  :extrapreds ((" << push << e[0]
             << space << e[1] << push << "))";
        }
        else {
          os << "  :extrafuns ((" << push << e[0]
             << space << e[1] << push << "))";
        }
      }
      else
	e.print(os);
      break;
    case SUBTYPE:
      throw SmtlibException("TheoryCore::print: SMTLIB: SUBTYPE not implemented");
      break;
    case TYPEDEF: {
      throw SmtlibException("TheoryCore::print: SMTLIB: TYPEDEF not implemented");
      break;
    }
    case EQ:
      // When a separator starts with a space (like " = "), add the
      // leading space with 'space' modifier.  If this separator goes
      // to the next line, the leading spaces must be eaten up to get
      // the indentation right; 'space' will tell the indentation
      // engine that it is a space that can be eaten.  A space in a
      // string (like " ") will never be eaten.
      os << "(" << push << "=" << space << e[0]
	 << space << e[1] << push << ")";
      break;
    case NOT:
      os << "(" << push << "not" << space << e[0] << push << ")";
      break;
    case AND: {
      int i=0, iend=e.arity();
      os << "(" << push << "and";
      for(; i!=iend; ++i) os << space << e[i];
      os << push << ")";
    }
      break;
    case OR: {
      int i=0, iend=e.arity();
      os << "(" << push << "or";
      for(; i!=iend; ++i) os << space << e[i] << space;
      os << push << ")";
      break;
    }
    case XOR: {
      int i=0, iend=e.arity();
      os << "(" << push << "xor";
      for(; i!=iend; ++i) os << space << e[i] << space;
      os << push << ")";
      break;
    }
    case ITE:
      os << "(" << push;
      if (getType(e) == boolType()) os << "if_then_else";
      else os << "ite";
      os << space << e[0]
	 << space << e[1] << space << e[2] << push << ")";
      break;
    case IFF:
      os << "(" << push << "iff" << space
	 << e[0] << space << e[1] << push << ")";
      break;
    case IMPLIES:
      os << "(" << push << "implies" << space
	 << e[0] << space << e[1] << push << ")";
      break;
      // Commands
    case ASSERT:
      throw SmtlibException("TheoryCore::print: SMTLIB: ASSERT not implemented");
      os << "(" << push << "ASSERT" << space << e[0] << push << ")";
      break;
    case TRANSFORM:
      throw SmtlibException("TheoryCore::print: SMTLIB: TRANSFORM not implemented");
      os << "(" << push << "TRANSFORM" << space << e[0] << push << ")";
      break;
    case QUERY:
      throw SmtlibException("TheoryCore::print: SMTLIB: QUERY not implemented");
      os << "(" << push << "QUERY" << space << e[0] << push << ")";
      break;
    case CONSTDEF:
    case LETDECL:
      //      throw SmtlibException("TheoryCore::print: SMTLIB: CONSTDEF, LETDECL not implemented");
      if(e.arity() == 2) os << e[1];
      else e.print(os);
      break;
    case LET: {
      // (LET ((var [ type ] val) .... ) body)
      for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i) {
	os << "(" << push;
        Type tp(getType(i->arity() == 3 ? (*i)[2] : (*i)[1]));
        DebugAssert(!tp.isNull(), "Unexpected Null type");
        if (tp.getExpr().getKind() == BOOLEAN) {
          os << "flet" << space << "(" << push;
        }
        else {
          os << "let" << space << "(" << push;
        }
	if(i->arity() == 3) {
	  os << (*i)[0] << space << nodag << (*i)[2];
	} else {
	  os << (*i)[0] << space << nodag << (*i)[1];
        }
	os << push << ")" << pop << pop << space;
      }
      os << e[1] << push;
      for (int j = 0; j < e[0].arity(); ++j)
        os << ")";
      break;
    }
    case BOUND_VAR:
      os << e.getName()+"_"+e.getUid();
      break;
    case PF_APPLY: {// FIXME: this will eventually go to the "symsim" theory
      throw SmtlibException("TheoryCore::print: SMTLIB: PF_APPLY not implemented");
      break;
    }
    case RAW_LIST: {
      throw SmtlibException("TheoryCore::print: SMTLIB: RAW_LIST not implemented");
      break;
    }
    case PUSH:
    case POP:
    case POPTO:
    case PUSH_SCOPE:
    case POP_SCOPE:
    case POPTO_SCOPE:
      throw SmtlibException("TheoryCore::print: SMTLIB: Stack operations not implemented");
      break;
    case PF_HOLE: // FIXME: implement this (now fall through to default)
    default:
      throw SmtlibException("TheoryCore::print: SMTLIB_LANG: Unexpected expression: "
                            +getEM()->getKindName(e.getKind()));
    }
    break; // end of case SMTLIB_LANG
  case LISP_LANG:
    switch(e.getKind()) {
    case TYPE:
      if(e.arity() == 0) os << "TYPE";
      else if(e.arity() == 1)
	os << "(" << push << "TYPE" << space << e[0] << push << ")";
      else if(e.arity() == 2)
	os << "(" << push << "TYPE" << space << e[0]
	   << space << e[1] << push << ")";
      else e.print(os);
      break;
    case ID: // FIXME: when lisp becomes case-insensitive, fix printing of IDs
      if(e.arity() == 1 && e[0].isString()) os << e[0].getString();
      else e.print(os);
      break;
    case BOOL_VAR:
      if(e.isBoolVar()) os << e.getVar();
      else e.print(os);
      break;
    case CONST:
      if(e.arity() == 2)
	os << "(" << push << "CONST" << space << e[0]
	   << space << e[1] << push << ")";
      else
	e.print(os);
      break;
    case SUBTYPE:
      os << "SUBTYPE(" << push << e[0] << push << ")";
      break;
    case TYPEDEF: {
      // This is used when dumping declarations to file.  Print just
      // the name of the type, unless it's a messed-up expression.
      if(e.arity() != 2) e.print(os);
      else if(e[0].isString()) os << e[0].getString();
      else e[0].print(os);
      break;
    }
    case EQ:
      // When a separator starts with a space (like " = "), add the
      // leading space with 'space' modifier.  If this separator goes
      // to the next line, the leading spaces must be eaten up to get
      // the indentation right; 'space' will tell the indentation
      // engine that it is a space that can be eaten.  A space in a
      // string (like " ") will never be eaten.
      os << "(" << push << "=" << space << e[0]
	 << space << e[1] << push << ")";
      break;
    case NOT:
      os << "(" << push << "NOT" << space << e[0] << push << ")";
      break;
    case AND: {
      int i=0, iend=e.arity();
      os << "(" << push << "AND";
      for(; i!=iend; ++i) os << space << e[i];
      os << push << ")";
    }
      break;
    case OR: {
      int i=0, iend=e.arity();
      os << "(" << push << "OR";
      for(; i!=iend; ++i) os << space << e[i] << space;
      os << push << ")";
    }
      break;
    case ITE:
      os << "(" << push << "IF" << space << e[0]
	 << space << e[1] << space << e[2] << push << ")";
      break;
    case IFF:
      os << "(" << push << "<=>" << space
	 << e[0] << space << e[1] << push << ")";
      break;
    case IMPLIES:
      os << "(" << push << "=>" << space
	 << e[0] << space << e[1] << push << ")";
      break;
      // Commands
    case ASSERT:
      os << "(" << push << "ASSERT" << space << e[0] << push << ")";
      break;
    case TRANSFORM:
      os << "(" << push << "TRANSFORM" << space << e[0] << push << ")";
      break;
    case QUERY:
      os << "(" << push << "QUERY" << space << e[0] << push << ")";
      break;
    case PUSH:
      os << "(PUSH)"; break;
    case POP:
      os << "(POP)"; break;
    case POPTO:
      os << "(" << push << "POPTO" << space << e[0] << push << ")"; break;
    case CONSTDEF:
    case LETDECL:
      if(e.arity() == 2) os << e[1];
      else e.print(os);
      break;
    case LET: {
      // (LET ((var [ type ] val) .... ) body)
      bool first(true);
      os << "(" << push << "LET" << space << "(" << push;
      for(Expr::iterator i=e[0].begin(), iend=e[0].end(); i!=iend; ++i) {
	if(!first) os << space;
	else first = false;
	os << "(" << push;
	if(i->arity() == 3) {
	  os << (*i)[0] << space << (*i)[1]
	     << space << nodag << (*i)[2];
	} else {
	  os << (*i)[0];
	  Type tp((*i)[0].lookupType());
	  if(!tp.isNull()) os << space << tp.getExpr();
	  os << space << nodag << (*i)[1];
	}
	os << push << ")" << pop << pop;
      }
      os << push << ")" << pop << pop << space << e[1] << push << ")";
      break;
    }
    case BOUND_VAR:
      os << e.getName()+"_"+e.getUid();
      break;
    case PF_APPLY: {// FIXME: this will eventually go to the "symsim" theory
      DebugAssert(e.arity() > 0, "TheoryCore::print(): "
		  "Proof rule application must have at "
		  "least one argument (rule name):\n "+e.toString());
      os << push << "(" << push;
      bool first(true);
      for(int i=0; i<e.arity(); i++) {
	if(first) first=false;
	else os << space;
	os << e[i];
      }
      os << push << ")";
      break;
    }
    case RAW_LIST: {
      os << "(" << push;
      bool firstTime(true);
      for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i) {
	if(firstTime) firstTime = false;
	else os << space;
	os << *i;
      }
      os << push << ")";
      break;
    }
    case PF_HOLE: // FIXME: implement this (now fall through to default)
    default:
      // Print the top node in the default LISP format, continue with
      // pretty-printing for children.
      e.print(os);
    }
    break; // end of case LISP_LANGUAGE
  default:
    // Print the top node in the default LISP format, continue with
    // pretty-printing for children.
    e.print(os);
  }
  return os;
}

///////////////////////////////////////////////////////////////////////////////
//parseExpr:
//translates intermediate representation to internal language for use in VCCmd
///////////////////////////////////////////////////////////////////////////////
Expr TheoryCore::parseExpr(const Expr& e) {
  TRACE("parseExpr", "parseExpr(", e, ") {");
  // Remember the current size of the bound variable stack
  size_t boundVarSize = d_boundVarStack.size();
  TRACE("parseExpr", "#bound vars = ", d_boundVarStack.size(), "");

  // Compute the kind to determine what to do with the expression
  int kind = NULL_KIND;

  switch(e.getKind()) {
  case ID: {
    TRACE("parser verbose", "resolving variable/operator(", e[0], ")");
    const Expr c1 = e[0];
    kind = getEM()->getKind(c1.getString());
    if(kind == NULL_KIND) { // It's an identifier; try to resolve it
      Expr res = resolveID(e[0].getString());
      if(res.isNull())
	throw ParserException("cannot resolve an identifier: "
			      +e.toString());
      else {
	TRACE("parseExpr", "#bound vars = ", d_boundVarStack.size(), "");
	TRACE("parseExpr", "parseExpr => ", res, " }");
	return res;
      }
    }
    // Otherwise exit the switch and use DP-specific parsing
    break;
  }
  case RAW_LIST: {
    TRACE("parser verbose", "constructing Expr from list(", e, ")");
    if(e.arity() == 0)
      throw ParserException("Empty list is not an expression!");
    const Expr& op = e[0];
    // First, resolve the operator
    switch(op.getKind()) {
    case ID: { // The operator in the list is ID: is it keyword or variable?
      kind = getEM()->getKind(op[0].getString());
      if(kind == NULL_KIND) {
	// It's a named function application.  Resolve the ID.
	Expr res = resolveID(op[0].getString());
	if(res.isNull())
	  throw ParserException("cannot resolve an identifier: "
				+e.toString());
	// Determine if it is an uninterpreted function or a lambda
	Expr op(res);
	// Kill the CONSTDEF and LETDECL
	while(op.getKind() == CONSTDEF || op.getKind() == LETDECL)
	  op = op[1];
	kind = APPLY; // Call TheoryUF
      }
      // Proceed to DP-specific parsing
      break;
    }
    case RAW_LIST: 
      // This can only be a lambda expression application.
      kind = LAMBDA;
      break;
    default:
      throw ParserException("Bad operator in "+e.toString());
    }
    break; // Exit the top-level switch, proceed to DP-specific parsing
  }
  default: // Can only be a string or a number.
    TRACE("parseExpr", "parseExpr => ", e, " }");
    return e;
  }

  // DP-specific parsing
  if(hasTheory(kind)) {
    TRACE("parseExpr", "parseExprOp["+theoryOf(kind)->getName()+"](",
	  e, ") {");
    Expr res = theoryOf(kind)->parseExprOp(e);
    TRACE("parseExpr",
	  "parseExprOp["+theoryOf(kind)->getName()+"] => ", res, " }");
    TRACE("parseExpr", "#bound vars = ", d_boundVarStack.size(), "");
    // Restore the bound variable stack
    while(d_boundVarStack.size() > boundVarSize) {
      TRACE("parseExpr", "Pop bound var "+d_boundVarStack.back().first+": ",
	    d_boundVarStack.back().second, "");
      d_boundVarStack.pop_back();
    }
    TRACE("parseExpr", "#bound vars = ", d_boundVarStack.size(), "");
    TRACE("parseExpr", "parseExpr => ", res, " }");
    return res;
  }
  else {
    TRACE("parseExpr", "parseExpr => ", e, " }");
    return e;
  }
}

//! An auxiliary recursive function to process COND expressions into ITE
Expr TheoryCore::processCond(const Expr& e, int i) {
  DebugAssert(i < e.arity()-1, "e = "+e.toString()+", i = "+int2string(i));
  if(i == e.arity()-2) {
    if(e[i].getKind() == RAW_LIST && e[i].arity() == 2
       && e[i+1].getKind() == RAW_LIST  && e[i+1].arity() == 2
       && e[i+1][0].getKind() == ID && e[i+1][0][0].getString() == "ELSE") {
      Expr c(parseExpr(e[i][0]));
      Expr e1(parseExpr(e[i][1]));
      Expr e2(parseExpr(e[i+1][1]));
      return c.iteExpr(e1,e2);
    }
  } else {
    if(e[i].getKind() == RAW_LIST && e[i].arity() == 2
       && e[i+1].getKind() == RAW_LIST  && e[i+1].arity() == 2) {
      Expr c(parseExpr(e[i][0]));
      Expr e1(parseExpr(e[i][1]));
      Expr e2(processCond(e, i+1));
      return c.iteExpr(e1,e2);
    }
  }
  throw ParserException("Parse Error: bad COND expression: "+e.toString());
}

///////////////////////////////////////////////////////////////////////////////
//parseExprOp:
//Recursive call of parseExpr defined in theory_ libaries based on kind of expr 
//being built
///////////////////////////////////////////////////////////////////////////////
Expr
TheoryCore::parseExprOp(const Expr& e) {
  TRACE("parser", "TheoryCore::parseExprOp(", e, ")");
  // If the expression is not a list, it must have been already
  // parsed, so just return it as is.
  switch(e.getKind()) {
  case ID: {
    int kind = getEM()->getKind(e[0].getString());
    switch(kind) {
    case NULL_KIND: return e; // nothing to do
    case TRUE:
    case FALSE:
    case TYPE:
    case BOOLEAN: return newExpr(getEM(), kind);
    default:
      DebugAssert(false, "Bad use of bare keyword: "+e.toString());
      return e;
    }
  }
  case RAW_LIST: break; // break out of switch, do the work
  default:
    return e;
  }
  DebugAssert(e.getKind()==RAW_LIST && e.arity() > 0 && e[0].getKind()==ID,
	      "TheoryCore::parseExprOp:\n e = "+e.toString());
  /* The first element of the list (e[0] is an ID of the operator. 
     ID string values are the dirst element of the expression */ 	      
  const Expr& c1 = e[0][0];
  int kind = getEM()->getKind(c1.getString());
  TRACE("parser verbose", "building kind(", kind, ")");
  switch(kind) {
  case VARDECLS: 
  case LETDECLS:
  case HELP: 
  case DUMP_PROOF:
  case DUMP_ASSUMPTIONS:
  case DUMP_SIG:
  case DUMP_TCC:
  case DUMP_TCC_ASSUMPTIONS:
  case DUMP_TCC_PROOF:
  case DUMP_CLOSURE:
  case DUMP_CLOSURE_PROOF:
  case COUNTEREXAMPLE:
  case WHERE:
  case ASSERT:
  case PRINT:
  case QUERY:
  case CHECKSAT:
  case CONTINUE:
  case RESTART:
  case TRACE:
  case ECHO:
  case UNTRACE:
  case VARLIST:
  case SUBTYPE: 
  case FORGET: 
  case GET_TYPE: 
  case IFF: 
  case IMPLIES: 
  case TYPEDEF: 
  case OPTION:
  case AND:
  case OR:
  case XOR:
  case NOT:
  case EQ: 
  case CALL: 
  case TRANSFORM: 
  case CHECK_TYPE: 
  case VARDECL: 
  case GET_CHILD: 
  case SUBSTITUTE: 
  case CONSTDEF: 
  case DBG: 
  case PUSH:
  case POP:
  case POPTO:
  case PUSH_SCOPE:
  case POP_SCOPE:
  case POPTO_SCOPE:
  case LETDECL:
  case ELSE:
  case CONTEXT:{
    vector<Expr> k;
    Expr::iterator i = e.begin(), iend=e.end();
    // Skip first element of the vector of kids in 'e'.
    // The first element is the operator.
    ++i; 
    // Parse the kids of e and push them into the vector 'k'
    for(; i!=iend; ++i) 
      k.push_back(parseExpr(*i));
    return newExpr(e.getEM(), kind, k);
  }
  case NEQ:
    if(e.arity()==3)
      return !(parseExpr(e[1]).eqExpr(parseExpr(e[2])));
    else
      throw ParserException("Disequality requires exactly two arguments: "
			    +e.toString());
    break;
  case CONST: {
    if(e.arity()==3) {
      vector<Expr> names;
      const Expr& vars = e[1];
      if(vars.getKind() == RAW_LIST) {
	for(Expr::iterator i=vars.begin(), iend=vars.end(); i!=iend; ++i)
	  names.push_back(*i);
      } else if(vars.getKind() == ID)
	names.push_back(vars);
      return newExpr(getEM(), CONST, newExpr(getEM(), RAW_LIST, names),
		     parseExpr(e[2]));
    }
    else if(e.arity() == 4 && e[1].getKind() == ID)
      return newExpr(getEM(), CONSTDEF, e[1],
		     parseExpr(e[2]), parseExpr(e[3]));
    throw ParserException("Bad CONST declaration: "+e.toString());
    break;
  }
  case TYPE: {
    if(e.arity()==2) {
      const Expr& types = e[1];
      if(types.getKind() == RAW_LIST) {
	vector<Expr> names;
	for(Expr::iterator i=types.begin(), iend=types.end(); i!=iend; ++i)
	  names.push_back(*i);
	return newExpr(getEM(), TYPEDECL, names);
      }
    }
    else if(e.arity() == 3 && e[1].getKind() == ID)
      return newExpr(getEM(), TYPEDEF, e[1], parseExpr(e[2]));
    throw ParserException("Bad TYPE declaration: "+e.toString());
    break;
  }
  case IF: 
    if(e.arity() == 4) {
      Expr c(parseExpr(e[1]));
      Expr e1(parseExpr(e[2]));
      Expr e2(parseExpr(e[3]));
      return c.iteExpr(e1, e2);
    } else
      throw ParserException("Bad IF-THEN-ELSE expression: "
		      +e.toString());
  case COND: {
    if(e.arity() >= 3)
      return processCond(e, 1);
    else
      throw ParserException("Bad COND expression: "+e.toString());
    break;
  }
  case LET: { // (LET ((v1 [ type1 ] e1) ... ) body)
    if(!(e.arity() == 3 && e[1].getKind() == RAW_LIST && e[1].arity() > 0))
      throw ParserException("Bad LET expression: "+e.toString());

    // Iterate through the bound variables
    for(Expr::iterator i=e[1].begin(), iend=e[1].end(); i!=iend; ++i) {
      const Expr& decl = *i;
      if(decl.getKind() != RAW_LIST
	 || !(decl.arity() == 2 || decl.arity() == 3))
	throw ParserException("Bad variable declaration block in LET "
			      "expression: "+decl.toString()+
			      "\n e = "+e.toString());
      if(decl[0].getKind() != ID)
	throw ParserException("Variable must be an identifier in LET "
			      "expression: "+decl[0].toString()+
			      "\n e = "+e.toString());
      Type type;
      Expr def;
      if(decl.arity()==3) {
	type = Type(parseExpr(decl[1]));
	def = parseExpr(decl[2]);
      } else 
	def = parseExpr(decl[1]);
      addBoundVar(decl[0][0].getString(), type, def);
    }
    // Parse the body recursively and return it (nuke the LET)
    return parseExpr(e[2]);
  }
//     {
//     vector<Expr> letDecls, lvars;
//     Expr letDeclsExpr;
//     for(Expr::iterator i = e[1].begin(), iend=e[1].end(); i!=iend; ++i) {
//       lvars = i->getKids();
//       lvars.insert(lvars.begin(), newExpr(e.getEM(), ID, newExpr(e.getEM(), STRING_EXPR, "LETDECL")));
//       letDecls.push_back(newExpr(e.getEM(), RAW_LIST, lvars));
//       }
//     letDecls.insert(letDecls.begin(), newExpr(e.getEM(), ID, newExpr(e.getEM(), STRING_EXPR, "LETDECLS")));
//     letDeclsExpr = newExpr(e.getEM(), RAW_LIST, letDecls);
//     return newExpr(e.getEM(), kind, parseExpr(letDeclsExpr), parseExpr(e[2]));
//   }
  case TRUE:	{ return e.getEM()->trueExpr();	}
  case FALSE:	{ return e.getEM()->falseExpr();}
  case BOOLEAN:	{ return e.getEM()->boolExpr();	}
    break;
  default:
    DebugAssert(false,
		"TheoryCore::parseExprOp: invalid command or expression: "
		+ e.toString());
    break;
  }
  return e;
}

///////////////////////////////////////////////////////////////////////////////
// Theory-cooperation framework code implementation                          //
///////////////////////////////////////////////////////////////////////////////


void TheoryCore::addFact(const Theorem& e)
{
  TRACE("facts addFact", "addFact[top level](", e, ") {");
  DebugAssert(!d_inAddFact, "Recursive call to addFact() is not allowed");
  DebugAssert(d_queue.empty(), "addFact[start]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "addFact[start]: Expected empty queue");
  IF_DEBUG(ScopeWatcher sw(&d_inAddFact));
  IF_DEBUG(if(*d_dumpTrace != "") debugger.dumpTrace(e.getExpr(), "addFact"));

  if(!d_inconsistent) enqueueFact(e);

  processFactQueue();

  DebugAssert(d_queue.empty(), "addFact[end]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "addFact[end]: Expected empty queue");

  TRACE_MSG("facts addFact", "addFact[top level] => }");
}


void TheoryCore::addEquality(const Theorem& e)
{
  TRACE("facts addFact", "addEquality[top level](", e, ") {");
  DebugAssert(!d_inAddFact,
	      "Recursive call to addFact()/addEquality() is not allowed");
  DebugAssert(d_queue.empty(), "addEquality[start]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "addEquality[start]: Expected empty queue");
  DebugAssert(e.isRewrite(), "addEquality("+e.getExpr().toString()+")");
  IF_DEBUG(ScopeWatcher sw(&d_inAddFact));
  IF_DEBUG(if(*d_dumpTrace != "")
	   debugger.dumpTrace(e.getExpr(), "addEquality"));

  if(!d_inconsistent) assertEqualities(e);

  processFactQueue();

  DebugAssert(d_queue.empty(), "addEquality[end]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "addEquality[end]: Expected empty queue");

  TRACE_MSG("facts addFact", "addEquality[top level] => }");
}


void TheoryCore::processFactQueue() {
  Theorem thm;
  int i;
  do {
    while (!d_queue.empty() && !d_inconsistent) {
      thm = d_queue.front();
      d_queue.pop();
      assertFactCore(thm);
    }
    if (d_inconsistent) {
      while(d_queue.size()) d_queue.pop();
      d_queueSE.clear();
      TRACE_MSG("facts addFact", "addFact[top level] => (inconsistent) }");
      return;
    }
    IF_DEBUG(debugger.counter("DP checkSAT calls")++);
    for(i=0; i<getNumTheories() && !d_inconsistent && d_queue.empty(); ++i) {
      TRACE("facts checkSat", "checkSat[", d_theories[i]->getName(), "]() {");
      d_theories[i]->checkSat(false);
      TRACE("facts checkSat", "checkSat[", d_theories[i]->getName(), "]() => }");
    }
    // Flush search engine queue inside the main loop; search engine
    // may add new facts to d_queue in the process.
    if(d_queue.empty()) {
      TRACE_MSG("facts addFact",
		"TheoryCore::processFactQueue(): flushing queueSE");
      // The queue may be emptied in the process due to
      // setInconsistent() call from the SearchEngine.  Monitor it at
      // every iteration.
      while(!d_queueSE.empty()) {
	// Copy a Theorem by value, to guarantee valid reference down
	// the call chain
	Theorem thm(d_queueSE.back());
	d_queueSE.pop_back();
	d_vcl->getSE()->addFact(thm);
      }
    }
  } while (!d_queue.empty());
}


bool TheoryCore::checkSATCore() {
  TRACE_MSG("facts checkSat", "checkSATCore() {");
  DebugAssert(d_queue.empty(), "checkSATCore[start]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "checkSATCore[start]: Expected empty queue");
  DebugAssert(!d_inCheckSATCore, "Recursive call to checkSATCore is not allowed!");
  IF_DEBUG(ScopeWatcher sw(&d_inCheckSATCore));
  IF_DEBUG(debugger.counter("DP checkSAT(fullEffort) calls")++);
  // Theory 0 is TheoryCore, skip it
  for(int i=0; i<getNumTheories() && !d_inconsistent && d_queue.empty(); ++i) {
    TRACE("facts checkSat", "checkSAT (full effort) [",
	  d_theories[i]->getName(), "]() {");
    d_theories[i]->checkSat(true);
    TRACE("facts checkSat", "checkSAT (full effort) [",
	  d_theories[i]->getName(), "]() }");
  }

  if(d_queue.empty()) {
    DebugAssert(d_queueSE.empty(),
		"checkSATCore[start]: Expected empty queue");
    TRACE_MSG("facts checkSat", "checkSATCore[queue empty] => true }");
    return true;
  } 

  // Process enqueued facts
  processFactQueue();
 
  TRACE("facts checkSat", "checkSATCore => ",
	d_inconsistent? "true" : "false", " }");

  DebugAssert(d_queue.empty(), "checkSATCore[start]: Expected empty queue");
  DebugAssert(d_queueSE.empty(), "checkSATCore[start]: Expected empty queue");

  return(d_inconsistent); 
}


void TheoryCore::addToVarDB(const Expr&  e)
{
  TRACE("model", "TheoryCore::addToVarDB(", e, ")");
  d_vars.push_back(e);
}

void TheoryCore::assertDisequalities() {
  // FIXME: disabled this function for an experiment
  return;
  TRACE("model", "Asserting disequalities of shared terms:",""  , "{ ");
  CDMap<Expr, Theory*>::iterator it = d_sharedTerms.begin(), it2,
    iend = d_sharedTerms.end();

  // First build a vector of theorems to assert, then actually assert
  // them.  Otherwise, asserting a new fact may change the set of
  // shared terms.
  vector<Theorem> diseqs;
  
  for(; it!=iend; ++it) {
    // Copy by value: the elements in the pair from *it are NOT refs in CDMap
    Expr e1 = (*it).first;
    DebugAssert(isTerm(e1), "TheoryCore::assertDisequalities(): "
		+e1.toString());
    for(it2 = it, ++it2; it2!=iend; ++it2) {
      Expr e2 = (*it2).first;
      Type tp(getBaseType(e1));
      if(findExpr(e1) != findExpr(e2) && tp == getBaseType(e2)) {
	Expr eq = e1.eqExpr(e2);
	Expr diseq = eq.notExpr();
	if(!d_vcl->getSE()->wasAsserted(diseq)) {
	  TRACE("model", "Disequality: [", diseq, "]");
	  diseqs.push_back(d_vcl->getSE()->newAssertion(diseq));
	}
      }
    }
  }
  for(vector<Theorem>::iterator i=diseqs.begin(), iend=diseqs.end();
      i!=iend; ++i) {
    addFact(*i);
  }
  TRACE("model", "Asserting disequalitites",""  , " }");
}


void TheoryCore::refineCounterExample()
{
  // Theory 0 is TheoryCore, skip it
  for(int i = 1; i<getNumTheories(); i++) {
    if(d_theories[i] != this)
      d_theories[i]->refineCounterExample();
    if(inconsistent()) {
      vector<Expr> assump;
      inconsistentThm().getAxiomAssumptions(assump);
      Expr a = d_vcl->listExpr(assump);
      throw EvalException("Theory["+d_theories[i]->getName()
		      +"]: Model Creation failed due "
		      "to the following assumptions:\n\n"
		      +a.toString()
		      +"\n\nYou might be using an incomplete logical fragment.");
    }
  }
}


void TheoryCore::computeModelBasic(const vector<Expr>& v) {
  for(vector<Expr>::const_iterator i=v.begin(), iend=v.end(); i!=iend; ++i) {
    TRACE("model", "Model var "+i->toString()+" = ", find(*i).getRHS(), "");
    DebugAssert(getType(*i).isBool(), "TheoryCore::computeModel: *i = "
		+(*i).toString());
    Expr val = find(*i).getRHS();
    if(!val.isBoolConst()) val = d_em->trueExpr();
    assignValue(*i, val);
  }
}


void TheoryCore::collectBasicVars() {
  TRACE_MSG("model", "collectBasicVars() {");
  // Clear caches
  d_varModelMap.clear();
  d_varAssignments.clear();
  // Current stack of variables to process
  vector<Expr> stack(d_vars.begin(), d_vars.end());
  size_t lastSize(0);
  while(stack.size() > 0) {
    Expr var(stack.back());
    stack.pop_back();
    if(d_varModelMap.count(var) > 0) continue; // Already processed
    Theorem findThm(find(var));
    if(findThm.getRHS()!=var) { // Replace var with its find
      d_simplifiedModelVars[var] = findThm;
      stack.push_back(findThm.getRHS());
      TRACE("model", "collectBasicVars: simplified var: ", findThm.getExpr(), "");
      continue; // Recycle to the beginning of the loop
    }
    lastSize = stack.size();
    TRACE("model", "getModelTerm(", var, ") {");
    getModelTerm(var, stack);
    TRACE("model", "getModelTerm => ",
	  newExpr(getEM(), RAW_LIST, stack), " }");
    if(stack.size() == lastSize) {
      // Add a primitive variable
      TRACE("model", "collectBasicVars: adding primitive var: ", var, "");
      d_basicModelVars.push_back(var);
      if(isTerm(var)) {
	Theory* t1 = theoryOf(var);
	Theory* t2 = theoryOf(getBaseType(var).getExpr().getKind());
	if(t1 != t2) {
	  TRACE("model", "collectBasicVars: adding shared var: ", var, "");
	  d_sharedTerms.insert(var, t1);
	  t1->addSharedTerm(var);
	  t2->addSharedTerm(var);
	}
      }
    } else { // Record the descendants of var
      vector<Expr>& kids = d_varModelMap[var];
      for(size_t i=lastSize; i<stack.size(); ++i) {
	DebugAssert(stack[i] != var, "Primitive var was pushed on "
		    "the stack in computeModelTerm(): "+var.toString());
	kids.push_back(stack[i]);
      }
    }
  }
  TRACE_MSG("model", "collectBasicVars() => }");
}


void TheoryCore::buildModel(ExprMap<Expr>& m) {
  TRACE_MSG("model", "buildModel() {");

  size_t numTheories = getNumTheories();
  // Use STL set to prune duplicate variables in theories
  set<Expr> theoryExprs[numTheories+1];
  // Sort out model vars by theories
  for(size_t j = 0 ; j<d_basicModelVars.size() ; j++) {
    const Expr& var = d_basicModelVars[j];
    Theorem findThm(find(var));
    if(findThm.getRHS()!=var) { // Replace var with its find and skip it
      d_simplifiedModelVars[var] = findThm;
      continue;
    }

    Theory *th = theoryOf(getBaseType(var).getExpr().getKind());
    size_t i=0;
    for(; i<numTheories && d_theories[i] != th; ++i);
    DebugAssert(d_theories[i] == th,
		"TheoryCore::buildModel: cannot find the theory ["
		+th->getName()+"] for the variable: "
		+var.toString());
    theoryExprs[i].insert(var);
    TRACE("model", "buildModel: adding ", var,
	  " to theory "+d_theories[i]->getName());
  }
  // Build a model for the basic-type variables
  for(int i=0; i<getNumTheories(); i++) {
    if(theoryExprs[i].size() > 0) {
      TRACE("model", "computeModelBasic[", d_theories[i]->getName(), "] {");
      // Copy the corresponding variables into a vector
      vector<Expr> vars;
      vars.insert(vars.end(), theoryExprs[i].begin(), theoryExprs[i].end());
      d_theories[i]->computeModelBasic(vars);
      TRACE("model", "computeModelBasic[", d_theories[i]->getName(), "] => }");
      if(inconsistent()) {
	vector<Expr> assump;
	inconsistentThm().getAxiomAssumptions(assump);
	Expr a = d_vcl->listExpr(assump);
	throw EvalException
	  ("Model Creation failed in Theory["
	   +d_theories[i]->getName()
	   +"] due to the following assumptions:\n\n"
	   +a.toString()
	   +"\n\nYou might be using an incomplete logical fragment.");
      }
    }
  }
  // Recombine the values for the compound-type variables
  ExprHashMap<Theorem>::iterator k, kend=d_simplifiedModelVars.end();
  for(CDList<Expr>::const_iterator i=d_vars.begin(), iend=d_vars.end(); i!=iend; ++i) {
    Expr var(*i);
    k=d_simplifiedModelVars.find(var);
    Theorem simp; // Null by default
    if(k!=kend) { // This var is simplified
      simp = k->second;
      var = simp.getRHS();
    }
    collectModelValues(var, d_simplifiedModelVars, m);
    if(d_varAssignments.count(var) > 0) {
      if(!simp.isNull()) {
	Theorem thm(transitivityRule(simp, d_varAssignments[var]));
	assignValue(thm);
	DebugAssert(thm.getLHS() == *i, "");
	m[*i] = thm.getRHS();
      } else
	m[*i] = d_varAssignments[var].getRHS();
    }
//     else if(simp.isNull())
//       m[*i] = *i;
//     else
//       m[*i] = simp.getRHS();
  }
  TRACE_MSG("model", "buildModel => }");
}


// Recursively build a compound-type variable assignment for e
/*! Not all theories may implement aggregation of compound values.  If
 *  a compound variable is not assigned by a theory, add the more
 *  primitive variable assignments to 'm'.
 *
 * Some variables may simplify to something else (simplifiedVars map).
 * INVARIANT: e is always simplified (it's not in simplifiedVars map).
 * Also, if v simplifies to e, and e is assigned, then v must be assigned.
 */
void TheoryCore::collectModelValues(const Expr& e,
				    const ExprHashMap<Theorem>& simplifiedVars,
				    ExprMap<Expr>& m) {
  TRACE("model", "collectModelValues(", e, ") {");
  if(d_varAssignments.count(e) > 0) {
    TRACE("model", "collectModelValues[cached] => ",
	  d_varAssignments[e].getRHS(), " }");
    return; // Got it already
  }
  if(d_varModelMap.count(e) == 0) { // Consider it a value of itself
    assignValue(reflexivityRule(e));
    TRACE("model", "collectModelValues[e=e] => ", e, " }");
    return; // Got it already
  }
  // Get the vector of more primitive vars
  const vector<Expr>& vars = d_varModelMap[e];
  // Recurse
  bool gotAll(true);  // Whether we got all the values
  ExprHashMap<Theorem>::iterator k, kend=simplifiedVars.end();
  for(vector<Expr>::const_iterator i=vars.begin(), iend=vars.end(); i!=iend; ++i) {
    Expr var(*i);
    k=simplifiedVars.find(var);
    Theorem simp; // Null by default
    if(k!=kend) { // This var is simplified
      simp = k->second;
      var = simp.getRHS();
    }
    collectModelValues(var, simplifiedVars, m);
    if(d_varAssignments.count(var) == 0) {
      gotAll = false;
    } else if(!simp.isNull()) {
      Theorem thm(transitivityRule(simp, d_varAssignments[var]));
      DebugAssert(thm.getLHS() == *i, "");
      assignValue(thm);
    }
  }
  IF_DEBUG(vector<Expr> res);
  if(gotAll) {
    vector<Expr> assigned; // What DP actually assigns
    Theory* th = theoryOf(getBaseType(e).getExpr());
    TRACE("model", "computeModel["+th->getName()+"](", e, ") {");
    th->computeModel(e, assigned);
    TRACE("model", "computeModel["+th->getName()+"] => ",
	  newExpr(getEM(), RAW_LIST, assigned), " }");
    // Check if the assigned vars are different from e
    if(!(assigned.size()==1 && assigned[0]==e)) {
      //    if(d_varAssignments.count(e) == 0) {
      for(vector<Expr>::iterator i=assigned.begin(), iend=assigned.end();
	  i!=iend; ++i) {
	if(*i == e) continue; // Skip the original var
	m[*i] = getModelValue(*i).getRHS();
	IF_DEBUG(res.push_back(getModelValue(*i).getExpr()));
      }
    } else {
      IF_DEBUG(res.push_back(getModelValue(e).getExpr()));
    }
  }
  TRACE("model", "collectModelValues => ",
	newExpr(getEM(), RAW_LIST, res), " }");
}


//! Test whether a kind maps to any theory
bool TheoryCore::hasTheory(int kind) {
  return (d_theoryMap.count(kind) > 0);
}


void TheoryCore::enqueueFact(const Theorem& e) {
  TRACE("enqueueFact", "enqueueFact[Core](", e.getExpr(), ")");
  // The theorem scope shouldn't be higher than current
  DebugAssert(e.getScope() <= d_vcl->scopeLevel(),
	      "\n e.getScope()="+int2string(e.getScope())
	      +"\n scopeLevel="+int2string(d_vcl->scopeLevel())
	      +"\n e = "+e.getExpr().toString());
  DebugAssert(d_inAddFact || d_inCheckSATCore,
	      "enqueueFact() is called outside of addFact()"
	      " or checkSATCore()");
  // No need to enqueue when already inconsistent
  if(inconsistent()) return;
	     
  IF_DEBUG(if(*d_dumpTrace != "")
	   debugger.dumpTrace(e.getExpr(), "enqueueFact"));
  // The theorem scope shouldn't be higher than current
  DebugAssert(e.getScope() <= d_vcl->scopeLevel(),
	      "\n e.getScope()="+int2string(e.getScope())
	      +"\n scopeLevel="+int2string(d_vcl->scopeLevel())
	      +"\n e = "+e.getExpr().toString());
  if(!e.isRewrite() && e.getExpr().isFalse()) {
    IF_DEBUG(debugger.counter("conflicts enqueued from DPs")++);
    setInconsistent(e);
  } else if(getResource()) {
    d_queue.push(e);
  } else {
    // No more resources: ignore all the remaining facts and fail
    // gracefully
    setIncomplete("Exhausted user-specified resource");
  }
}


void TheoryCore::enqueueEquality(const Theorem& e) {
  TRACE("enqueueEquality", "enqueueEquality["+getName()+"](",
	e.getExpr(), ")");
  IF_DEBUG(if(*d_dumpTrace != "")
	   debugger.dumpTrace(e.getExpr(), "enqueueEquality"));
  IF_DEBUG(const Expr& expr = e.getExpr());
  IF_DEBUG(int kind = expr.getKind());
  DebugAssert(kind == EQ || kind == AND || kind == EXISTS || kind == FALSE,
	      "enqueueEnquality: bad input: "+expr.toString());
  // The theorem scope shouldn't be higher than current
  DebugAssert(e.getScope() <= d_vcl->scopeLevel(),
	      "\n e.getScope()="+int2string(e.getScope())
	      +"\n scopeLevel="+int2string(d_vcl->scopeLevel())
	      +"\n e = "+expr.toString());
//   DebugAssert(kind != EQ || !expr[0].subExprOf(expr[1]),
// 	      "enqueueEquality("+expr.toString()+")");
  d_equalityQueue.push_back(e);
}


void TheoryCore::addSplitter(const Expr& e, int priority) {
  d_vcl->getSE()->addSplitter(e, priority);
}


void TheoryCore::assertFactCore(const Theorem& e)
{
  TRACE("facts assertFact", "assertFactCore(", e, ") {");
  const Expr& e2 = e.getExpr();
  Theorem equiv = simplifyThm(e2);

  Theorem estarThm(d_commonRules->iffMP(e, equiv));
  Expr estar = estarThm.getExpr();

  if(estar.isExists()) {
    Theorem sk(skolemize(estarThm));
    enqueueFact(sk);
    TRACE("skolem", "assertFactCore: created Skolemized assumption:\n ",
	  sk.getExpr(), "");
  }

  // Transform disequalities using structural similarities, e.g.:
  // f(x,y) != f(z,y) becomes x != z.  New disequalities are enqueued,
  // and the original is processed (for completeness).
  if(estar.isNot() && d_vcl->getFlags()["diseq"].getBool()) {
    int kind(estar[0].getKind());
    if(kind==IFF || kind==EQ) {
      Theorem thm = transformDiseq(estarThm);
      if(thm != estarThm) enqueueFact(thm);
    }
  }

  // Make sure originally asserted atomic formulas have a find pointer
  if(e2 != estar && !e2.isTrue() && isLiteral(e2))
    setFindLiteral(e, true);
  
  if(estar.isAnd()) {
    for(int i=0,iend=estar.arity(); i<iend; ++i)
      assertFactCore(andElim(estarThm, i));
    TRACE_MSG("facts assertFact", "assertFactCore[AND] => }");
    return;
  }

  if (isLiteral(estar)) {
    if (estar.isEq()) {
      // Notify the search engine
      enqueueSE(estarThm);
      Theorem solvedThm(solve(estarThm));
      // Search engine doesn't have to know about estar at this point
      if(estar != solvedThm.getExpr()) setFindLiteral(estarThm, false);
      assertEqualities(solvedThm);
    }
    else if (estar.isFalse()) {
      IF_DEBUG(debugger.counter("conflicts from simplifier")++);
      setInconsistent(estarThm);
    }
    else if (!estar.isTrue()) {
      assertFormula(estarThm);
    }
  } else {
    // Notify the search engine
    enqueueSE(estarThm);
  }

  DebugAssert(!isLiteral(e2) || e2.hasFind()
	      || (e2.isNot() && e2[0].hasFind()),
	      "assertFactCore: e2 = "+e2.toString());
  DebugAssert(!isLiteral(estar) || estar.hasFind()
	      || (estar.isNot() && estar[0].hasFind()),
	      "assertFactCore: estar = "+estar.toString());
  TRACE_MSG("facts assertFact", "assertFactCore => }");
}

/*!
 * Fill in the copy of the equality queue with single equalities by
 * processing the input 'eq', which can be an equality, a conjunction
 * of equalities, or an existential quantifier of the above.
 */
bool TheoryCore::processEquality(const Theorem& thm, ExprMap<Theorem>& q) {
  const Expr& e = thm.getExpr();
  TRACE("facts eq", "processEquality(", e, ") {");
  bool res(false);

  switch(e.getKind()) {
  case FALSE:
    IF_DEBUG(debugger.counter("conflicts from solver")++);
    setInconsistent(thm);
    res = true;
    break;
  case EXISTS:
    // Skolemize the quantifier and expect an equality or a
    // conjunction of equalities inside
    res = processEquality(skolemize(thm), q);
    break;
  case AND: 
    for(int i=0, iend=e.arity(); i!=iend; ++i)
      if(res = processEquality(andElim(thm, i), q)) break;
    break;
  case EQ:
    // find(e0) != e1 && find(e1) != e0
    if(find(e[0]).getRHS() != e[1] && e[0] != find(e[1]).getRHS()) {
      DebugAssert(e[0] != e[1], "processEquality: e = "+e.toString());
      DebugAssert(!e[1].hasFind() || find(e[1]).getRHS() == e[1],
		  "processEquality: e = "+e.toString()
		  +"\nfind(e[1]) = "+find(e[1]).getRHS().toString());
//       DebugAssert(!e[0].subExprOf(e[1]), "processEquality: e = "+e.toString());
      q[e] = thm;
      TRACE("facts eq", "  [new]  ", e, ",\n");
    } else if(find(e).getRHS().isFalse()) {
      TRACE("facts eq", " [FALSE] ", e,  "]");
      IF_DEBUG(debugger.counter("conflicts from find pointers")++);
      setInconsistent(contradictionRule(thm, iffFalseElim(find(e))));
      res = true;
    } else {
      TRACE("facts eq", "         ", e, ",");
    }
    break;
  default:
    DebugAssert(false,
		"TheoryCore::processEquality(): bad input:\n "
		+thm.toString());
  }
  TRACE("facts eq", "processEquality => ", res? "true" : "false", " }");
  return res;
}

/*!
 * Traverses lhs and rhs in parallel, till they become different.
 * Once the top nodes of lhs and rhs become different, add an equality
 * lhs=rhs to the 'eqs' vector.
 *
 * \return a proof of lhs=rhs under the assumptions from eqs.
 */
Theorem 
TheoryCore::traverseDiseq(const Expr& lhs, const Expr& rhs,
			  ExprMap<Theorem>& cache, std::vector<Expr>& eqs) {
  // First, check the trivial case
  DebugAssert(lhs!=rhs, "traverseDiseq("+lhs.toString()+")");
  // Then search the cache
  bool useEq(isTerm(lhs));
  Expr eq(useEq? lhs.eqExpr(rhs) : lhs.iffExpr(rhs));
  ExprMap<Theorem>::iterator i=cache.find(eq), iend=cache.end();
  if(i!=iend) return (*i).second;
  if((i=cache.find(useEq? rhs.eqExpr(lhs) : rhs.iffExpr(lhs)))!=iend)
    return symmetryRule((*i).second);

  // Equation is not in the cache, do the work
  Theorem res;
  bool kidsMatch(lhs.arity() == rhs.arity() && getOp(lhs) == getOp(rhs));
  if(kidsMatch) { // Now check that the types of kids also match
    for(int i=0,iend=lhs.arity(); kidsMatch && i<iend; ++i)
      kidsMatch = (getBaseType(lhs[i]) == getBaseType(rhs[i]));
  }
  if(!kidsMatch) {
    // Top nodes are different, create a new equation
    eqs.push_back(eq);
    res=assumpRule(eq);
  } else {
    // Top nodes are the same, recurse
    vector<Theorem> thms;
    vector<unsigned> changed;
    for(int j=0,jend=lhs.arity(); j<jend; ++j) {
      if(lhs[j]!=rhs[j]) {
	thms.push_back(traverseDiseq(lhs[j], rhs[j], cache, eqs));
	changed.push_back(j);
      }
    }
    DebugAssert(changed.size()>0, "traverseDiseq("+lhs.toString()+", "
		+rhs.toString()+")");
    res = substitutivityRule(lhs, changed, thms);
  }
  cache[eq] = res;
  return res;
}


Theorem
TheoryCore::transformDiseq(const Theorem& thm) {
  TRACE("facts diseq", "transformDiseq(", thm.getExpr(), ") {");
  DebugAssert(thm.getExpr().isNot(), "transformDiseq("+thm.toString()+")");
  const Expr& eq = thm.getExpr()[0];
  DebugAssert(eq.isEq() || eq.isIff(), "transformDiseq("+thm.toString()+")");
  const Expr& lhs = eq[0];
  const Expr& rhs = eq[1];
  DebugAssert(lhs != rhs, "transformDiseq("+thm.getExpr().toString()+")");
  bool kidsMatch(lhs.arity() == rhs.arity() && getOp(lhs) == getOp(rhs));
  if(kidsMatch) { // Now check that the types of kids also match
    for(int i=0,iend=lhs.arity(); kidsMatch && i<iend; ++i)
      kidsMatch = (getBaseType(lhs[i]) == getBaseType(rhs[i]));
  }
  if(!kidsMatch) {
    TRACE("facts diseq", "transformDiseq[no change] => ", thm.getExpr(), " }");
    return thm;
  }
  // Traverse the expressions recursively and collect the different
  // subtrees.  The resulting theorem is (x1=y1&...&xn=yn)|-lhs=rhs,
  // from which we derive the contrapositive, and the disjunction of
  // disequalities.
  vector<Expr> eqs;
  ExprMap<Theorem> cache;
  Theorem res = traverseDiseq(lhs, rhs, cache, eqs);
  // Derive the implication
  res = d_commonRules->implIntro(res, eqs);
  // Then derive the contrapositive and apply modus ponens
  res = d_commonRules->implMP(thm, d_commonRules->implContrapositive(res));
  // Now propagate the negation down
  res = iffMP(res, pushNegation(res.getExpr()));
  // And simplify the result
  res = iffMP(res, simplifyThm(res.getExpr()));
  IF_DEBUG(debugger.counter("diseq expanded")++);
  TRACE("facts diseq", "transformDiseq => ", res.getExpr(), " }");
  return res;
}


/*!
 * Invariants: 
 * - The input theorem e is either an equality or a conjunction of
 *   equalities;
 * - In each equality e1=e2, the RHS e2 i-leaf-simplified;
 * - At the time of calling update(), all equalities in the queue are
 *   processed by assertFormula() and the equivalence classes are merged
 *   in the union-find database.
 *
 * In other words, the entire equality queue is processed "in parallel".
 */

void TheoryCore::assertEqualities(const Theorem& e)
{
  TRACE("facts eq", "assertEqualities(", e, ") {");
  DebugAssert(d_equalityQueue.empty(), "TheoryCore::assertEqualities()");
  enqueueEquality(e);

  while(!d_equalityQueue.empty()) {
    // We need to be careful: first copy the queue to a separate map
    // (to eliminate duplicates) and empty the "official" queue, so
    // that DPs can safely enqueue new equalities.

    ExprMap<Theorem> q;
    for(vector<Theorem>::iterator i=d_equalityQueue.begin(),
	  iend=d_equalityQueue.end(); i!=iend; ++i) {
      if(processEquality(*i, q)) {
	// Derived inconsistency
	TRACE_MSG("facts eq", "assertEqualities[FALSE] => }");
	d_equalityQueue.clear();
	return;
      }
    }
    d_equalityQueue.clear();

    // Check invariants and assert the equations to the DPs
    for(ExprMap<Theorem>::iterator i=q.begin(), iend=q.end(); i!=iend; ++i) {
      const Theorem& e = i->second;
      DebugAssert(e.isRewrite(), "");
      DebugAssert(isTerm(e.getLHS()), "term expected");

      assertFormula(e);
    }

    // Merge the equivalence classes
    for(ExprMap<Theorem>::iterator i=q.begin(), iend=q.end(); i!=iend; ++i) {
      const Theorem& e = i->second;
      DebugAssert(find(e.getLHS()).getRHS() == e.getLHS(),
		  "find error: e = "+e.getExpr().toString()
		  +"\n\n  find(e.getLHS()) = "
		  +find(e.getLHS()).getRHS().toString());
      DebugAssert(find(e.getRHS()).getRHS() == e.getRHS(),
		  "find error: e = "+e.getExpr().toString()
		  +"\n\n  find(e.getRHS()) = "
		  +find(e.getRHS()).getRHS().toString());
//       DebugAssert(leavesAreSimp(e.getRHS()), "assertEqualities("
// 		  +e.toString()+")");
      
      e.getLHS().setFind(e);
    }

    d_em->invalidateSimpCache();
  
    // Call the update() functions (process NotifyLists).  

    for(ExprMap<Theorem>::iterator i=q.begin(), iend=q.end(); i!=iend; ++i) {
      const Theorem& e = i->second;
      NotifyList *L = e.getLHS().getNotify();
      if (L) {
	// L may be changing during update() call; copy it to a local
	// vector before calling update()
	vector<pair<Theory*, Expr> > l;
	for (unsigned k = 0, size = L->size(); k < size; ++k) {
	  l.push_back(pair<Theory*,Expr>(L->getTheory(k), L->getExpr(k)));
	}
	for(size_t k = 0, size = l.size(); k < size; ++k) {
	  TRACE("facts update", "update["+l[k].first->getName()+"]("
		+e.toString()+", ", l[k].second, ") {");
	  l[k].first->update(e, l[k].second);
	  TRACE("facts update", "update[", l[k].first->getName(), "] => }");
	}
      }
    }
  }
  TRACE_MSG("facts eq", "assertEqualities => }");
}


void TheoryCore::assertFormula(const Theorem& thm)
{
  TRACE("facts assertFact", "assertFormula(", thm, ") {");
  const Expr& e = thm.getExpr();
  if (isLiteral(e)) {
    Theory* i = theoryOf(e);
    // (Dis)equalities should also be asserted to the theories of
    // their types for concrete model generation 
    Theory* i2(NULL);
    // Treat disequality as a separate theory, so that terms are
    // reported as shared
    if(e.isNot() && e[0].isEq())
      setupTerm(e, this);
    else
      setupTerm(e,i);

    // Use find to force af to rewrite to TRUE and NOT af to rewrite to FALSE,
    // where af is an atomic formula.  If af is an equality, make sure its lhs
    // is greater than its rhs so the simplifier will be able to use the find.
    enqueueSE(thm);
    if (e.isNot()) {
      DebugAssert(!e[0].hasFind(), "already has find");
      if (e[0].isEq()) {
	// Save the disequality for later processing
	d_diseq.push_back(thm);
	// For concrete model generation
	i2 = theoryOf(getBaseType(e[0][0]).getExpr()); 

	if(e[0][0] < e[0][1]) {
	  Expr e2 = e[0][1].eqExpr(e[0][0]);
	  DebugAssert(!e2.hasFind(), "already has find");
	  e2.setFind(transitivityRule(rewriteUsingSymmetry(e2),
				      notToIff(thm)));
	}
      }
      e[0].setFind(notToIff(thm));
    }
    else {
      // If we are called from assertEqualities(), we may get e1=e2
      // for which there is a known fact e1/=e2 or e2/=e1.  In this
      // case, we should derive a contradiction.
      // Note, that this is not necessarily detected by the DPs.
      if(e.hasFind()) {
	// Why would find(e) be always false?  Not if we set find to
	// all atomic formulas...
// 	DebugAssert(find(e).getRHS().isFalse(), "e = "+e.toString()
// 		    +"\nfind(e) = "+find(e).getRHS().toString());
	Theorem findThm(find(e));
	if(findThm.getExpr().isFalse()) {
	  IF_DEBUG(debugger.counter("conflicts from find pointers")++);
	  setInconsistent(contradictionRule(thm, iffFalseElim(find(e))));
	  TRACE_MSG("facts assertFact", "assertFormula[contradiction] => }");
	  return;
	}
      }
      else
	e.setFind(d_commonRules->iffTrue(thm));

      if(e.isEq()) {
	if(e[0] < e[1]) {
	  Expr e2 = e[1].eqExpr(e[0]);
	  if(e2.hasFind()) {
// 	    DebugAssert(find(e2).getRHS().isFalse(), "e2 = "+e2.toString()
// 			+"\nfind(e2) = "+find(e2).getRHS().toString());
	    Theorem findThm(find(e2));
	    if(findThm.getExpr().isFalse()) {
	      IF_DEBUG(debugger.counter("conflicts from find pointers")++);
	      setInconsistent(contradictionRule(symmetryRule(thm),
						iffFalseElim(find(e2))));
	      TRACE_MSG("facts assertFact", "assertFormula[e[1]/=e[0]] => }");
	      return;
	    }
	  } else
	    e2.setFind(transitivityRule(rewriteUsingSymmetry(e2),
					d_commonRules->iffTrue(thm)));
	}
      }
    }
    d_em->invalidateSimpCache();
    // Do not send existential quantifiers to DPs; skolemize them and
    // notify the Search Engine about it.
    if(thm.getExpr().isExists())
      enqueueFact(skolemize(thm));
    else {
      TRACE("facts assertFact", ("assertFact[->" + i->getName() + "]("), thm, ") {");
      i->assertFact(thm);
      TRACE("facts assertFact", "assertFact[->", i->getName(),"] => }");
      if(i2 != NULL && i != i2) {
	TRACE("facts assertFact", ("assertFact[->" + i2->getName() + "]("), thm, ") {");
	IF_DEBUG(debugger.counter("assertFact(diseq) by type")++);
	i2->assertFact(thm);
	TRACE("facts assertFact", "assertFact[->", i2->getName(),"] => }");
      }
    }
  }
  else {
    DebugAssert(false,"assertFormula: nonliteral asserted:\n  "
		+thm.toString());
  }
  TRACE_MSG("facts assertFact", "assertFormula => }");
}


void TheoryCore::setInconsistent(const Theorem& e) {
  TRACE("setInconsistent", "setInconsistent(", e, ") {");
  d_inconsistent = true; d_incThm = e;
  d_queueSE.clear();
  // Theory 0 is TheoryCore, skip it
  for(int i=1; i<getNumTheories(); ++i) {
    TRACE("facts notifyInconsistent", "notifyInconsistent[", d_theories[i]->getName(), "]() {");
    d_theories[i]->notifyInconsistent(e);
    TRACE("facts notifyInconsistent", "notifyInconsistent[", d_theories[i]->getName(), "]() => }");
  }
  TRACE_MSG("setInconsistent", "setInconsistent => }");
}


void TheoryCore::setIncomplete(const string& reason) {
  d_incomplete.insert(reason, true);
}


void TheoryCore::setupTerm(const Expr& t, Theory* i)
{
  TRACE("facts setup", "setupTerm(", t, ", i="+i->getName()+") {");

  // Even if t is already setup, it may become shared with another theory
  Theory* j = theoryOf(t);
  if(i != j && isTerm(t)) {
    TRACE("facts setup", "addSharedTerm[->"+j->getName()+"](", t, ") {");
    j->addSharedTerm(t);
    TRACE_MSG("facts setup", "addSharedTerm[->"+j->getName()+"] => }");
    TRACE("facts setup", "addSharedTerm[->"+i->getName()+"](", t, ") {");
    i->addSharedTerm(t);
    TRACE_MSG("facts setup", "addSharedTerm[->"+i->getName()+"] => }");
    d_sharedTerms.insert(t, j);
  }
  
  int k;

  // Atomic formulas (non-terms) may have find pointers without the
  // subterms being setup.  Recurse down to the terms before checking
  // for find pointers.
  if (!isTerm(t)) {
    if(isLiteral(t)) {
      for (k = 0; k < t.arity(); ++k) {
	setupTerm(t[k], i);
      }
      TRACE("facts setup", "setup[->"+j->getName()+"](", t, ") {");
      j->setup(t);
      TRACE("facts setup", "setup[->",j->getName(),"] => }");
    }
    TRACE_MSG("facts setup", "setupTerm[not term] => }");
    return;
  }

  // If already setup, nothing else to do
  if (t.hasFind()) {
    TRACE_MSG("facts setup", "setupTerm[has find] => }");
    return;
  }

  // Proceed with the setup

  // Add the term into the set of all terms
  d_terms.push_back(t);

  for (k = 0; k < t.arity(); ++k) {
    setupTerm(t[k],theoryOf(t));
  }
  Theorem thm = d_commonRules->reflexivityRule(t);
  t.setFind(thm);
  TRACE("facts setup", "setup[->"+j->getName()+"](", t, ") {");
  j->setup(t);
  TRACE("facts setup", "setup[->",j->getName(),"] => }");

  // Assert the subtyping predicate AFTER the setup is complete
  Theorem pred = typePred(t);
  // pred = d_commonRules->iffMP(pred, simplifyThm(pred.getExpr()));
  const Expr& predExpr = pred.getExpr();
  // TRACE("tccs", "setupTerm: subtype predicate:\n", predExpr, "");
  if(predExpr.isFalse()) {
    IF_DEBUG(debugger.counter("conflicts from type preficate")++);
    setInconsistent(pred); 
  }
  else if(!predExpr.isTrue()) {
    Theory* k = theoryOf(getType(t).getExpr());
    TRACE("facts setup", "assertTypePred[->"+k->getName()+"]("
	  +t.toString()+", ", pred, ") {");
    k->assertTypePred(t, pred);
    TRACE_MSG("facts setup", "assertTypePred[->"+k->getName()+"] => }");
  }

  TRACE_MSG("facts setup", "setupTerm => }");
}

void
TheoryCore::setResourceLimit(unsigned limit) {
  d_resourceLimit = limit;
}


IF_DEBUG(static ExprMap<bool> simpStack);

Theorem TheoryCore::simplifyThm(const Expr& e, bool forceRebuild)
{
  TRACE("simplify", "simplifyThm(", e, ") {");
  IF_DEBUG(ScopeWatcher sw(&d_inSimplify));

  if (*d_simplifyInPlace) {
    if (!forceRebuild) {
      Theorem res(simplifyInPlaceThmRec(e));
      TRACE("simplify", "simplifyThm[in place] => ", res.getRHS(), " }");
      return res;
    }
    else {
      d_em->invalidateSimpCache();
      Theorem thm = simplify(e);
      d_em->invalidateSimpCache();
      if (!isLiteral(e) && e != thm.getRHS()) {
	e.setFind(thm);
      }
      TRACE("simplify", "simplifyThm[rebuild] => ", thm.getRHS(), " }");
      return thm;
    }
  }
  Theorem res(simplify(e));
  TRACE("simplify", "simplifyThm => ", res.getRHS(), " }");
  return res;
}

Theorem TheoryCore::simplify(const Expr& e)
{
  TRACE("simplifyRec", "simplify(", e, ") {");
  DebugAssert(simpStack.count(e) == 0, "TheoryCore::simplify: loop detected over e =\n"
	      +e.toString());
  TRACE("simplify stack", "simpStack.size = ", simpStack.size(), "");
  DebugAssert(simpStack.size() < 10000,
	      "TheoryCore::simplify: too deep recursion depth");
  IF_DEBUG(simpStack[e] = true);
  Theorem thm;
  if(e.hasFind()) {
    thm = find(e);
    const Expr& e2 = thm.getRHS();
    if (e2.hasFind()) {
      DebugAssert(e2.isTrue() || e2.isFalse() || isTerm(e2),
		  "Unexpected find pointer");
      TRACE_MSG("simplifyRec", "Using find");
      TRACE("simplifyRec", "simplify => ", thm, "}\n");
      IF_DEBUG(simpStack.erase(e));
      return thm;
    }
    else {
      DebugAssert(!isLiteral(e), "Expected non-literal: e = "+e.toString()
		  +"\n e2 = "+e2.toString());
      DebugAssert(*d_simplifyInPlace, "Expected in-place flag: e = "
		  +e.toString()+"\n e2 = "+e2.toString());
      Theorem thm2 = simplify(e2);
      if (thm2.getLHS() != thm2.getRHS()) {
	e.setFind(transitivityRule(thm,thm2));
	TRACE_MSG("simplifyRec", "Using find 2");
	TRACE("simplifyRec", "simplify => ", e.getFind(), "}\n");
	IF_DEBUG(simpStack.erase(e));
	return e.getFind();
      }
      TRACE_MSG("simplifyRec", "Using find 3");
      TRACE("simplifyRec", "simplify => ", thm, "}\n");
      IF_DEBUG(simpStack.erase(e));
      return thm;
    }
  }

  if(e.validSimpCache()) {
    TRACE_MSG("simplifyRec", "Using cached result");
    TRACE("simplifyRec", "simplify => ", e.getSimpCache(), "}\n");
    IF_DEBUG(simpStack.erase(e));
    return e.getSimpCache();
  }

  // Optimization: check both sides of = or <=> for syntactic equality
  switch(e.getKind()) {
  case EQ:
  case IFF:
    if(e[0]==e[1]) {
      thm = iffTrue(reflexivityRule(e[0]));
      IF_DEBUG(debugger.counter("simplified x=x")++);
      break; // Break out of switch()
    }
    // Otherwise fall through
  default: // Make a recursive call
    thm = theoryOf(e)->simplifyOp(e);
    thm = rewriteCore(thm);
  }

  const Expr& e2 = thm.getRHS();
#ifdef DEBUG
    if (!isTerm(e2) || !e2.hasFind()) {
      // The rewriter should guarantee that all of its children are simplified.
      for (int k=0; k<e2.arity(); ++k) {
	Expr simplified(simplify(e2[k]).getRHS());
	DebugAssert(e2[k]==simplified,"Simplify Error 1:\n e2[k="+int2string(k)
		    +"] = "
		    +e2[k].toString() + "\nSimplified = "
		    +simplified.toString()
		    +"\ne2 = "+e2.toString());
// 	Expr rewritten(rewriteCore(e2[k]).getRHS());
// 	DebugAssert(e2[k]==rewritten,"Simplify Error 2: e2[k] = \n"
// 		    +e2[k].toString() + "\nSimplified rewritten = \n"
// 		    +rewritten.toString());
      }
    }
    Expr rewritten(rewriteCore(e2).getRHS());
    DebugAssert(e2==rewritten,"Simplify Error 2: e2 = \n"
		+e2.toString() + "\nSimplified rewritten = \n"
		+rewritten.toString());
#endif
  e.setSimpCache(thm);
  if (e != e2) {
    e2.setSimpCache(reflexivityRule(e2));
  }
  TRACE("simplifyRec", "simplify => ", thm, "}\n");
  IF_DEBUG(simpStack.erase(e));
  return thm;
}


Theorem TheoryCore::simplifyInPlaceThmRec(const Expr& e)
{
  TRACE("simplifyThmRec", "simplifyInPlaceThmRec(", e, ") {");
  Theorem thm;
  if (e.hasFind()) {
    thm = find(e);
    const Expr& e2 = thm.getRHS();
    if (e2.hasFind()) {
      DebugAssert(e2.isTrue() || e2.isFalse() || isTerm(e2),
		  "Unexpected find pointer");
      TRACE_MSG("simplifyThmRec", "Using find");
      TRACE("simplifyThmRec", "simplifyInPlaceThmRec => ", thm, "}\n");
      return thm;
    }
    else {
      DebugAssert(!isLiteral(e), "huh?");
      Theorem thm2 = simplifyInPlaceThmRec(e2);
      if (thm2.getLHS() != thm2.getRHS()) {
	e.setFind(transitivityRule(thm,thm2));
	TRACE_MSG("simplifyThmRec", "Using find 2");
	TRACE("simplifyThmRec", "simplifyInPlaceThmRec => ", e.getFind(), "}\n");
	return e.getFind();
      }
      TRACE_MSG("simplifyThmRec", "Using find 3");
      TRACE("simplifyThmRec", "simplifyInPlaceThmRec => ", thm, "}\n");
      return thm;
    }
  }

  if (e.validSimpCache()) {
    TRACE_MSG("simplifyThmRec", "Using cached result");
    TRACE("simplifyThmRec", "simplifyInPlaceThmRec => ", e.getSimpCache(), "}\n");
    return e.getSimpCache();
  }

  int k, ar = e.arity();
  if (ar > 0) {
    vector<Theorem> newChildrenThm;
    vector<unsigned> changed;
    const vector<Expr>& kids = e.getKids();
    for(k = 0; k < ar; ++k) {
      thm = simplifyInPlaceThmRec(kids[k]);
      if (thm.getLHS() != thm.getRHS()) {
        if ((e.getKind() == OR && thm.getRHS().isTrue()) ||
            (e.getKind() == AND && thm.getRHS().isFalse())) {
          newChildrenThm.clear();
          changed.clear();
          newChildrenThm.push_back(thm);
          changed.push_back(k);
          break;
        }
        else if (k == 0 && e.getKind() == ITE && (thm.getRHS().isBoolConst())) {
          newChildrenThm.push_back(thm);
          changed.push_back(k);
          if (thm.getRHS().isTrue()) k += 1;
          else k += 2;
          thm = simplify(kids[k]);
          newChildrenThm.push_back(thm);
          changed.push_back(k);          
          break;
        }
        else {
          newChildrenThm.push_back(thm);
          changed.push_back(k);
        }
      }
    }
    if (changed.size() == 0) thm = rewriteCore(e);
    else {
      thm = substitutivityRule(e, changed, newChildrenThm);
      Theorem thm2 = rewriteCore(thm.getRHS());
      const Expr& e2 = thm2.getLHS();
      if (e2 == thm2.getRHS()) {
	if (!isLiteral(e) && !isAtomic(e2) && !isLiteral(e2)) {
	  e.setFind(thm);
	  thm = reflexivityRule(e);
	}
      }
      else thm = transitivityRule(thm, thm2);
    }
  } else {
    thm = rewriteCore(e);
  }
  e.setSimpCache(thm);
  const Expr& e2 = thm.getRHS();
  if (e != e2) {
    e2.setSimpCache(reflexivityRule(e2));
  }
  TRACE("simplifyThmRec", "simplifyInPlaceThmRec => ", thm, "}\n");
  return thm;
}


#ifdef DEBUG
static void checkRewriteType(TheoryCore* core, const Theorem& thm) {
  Type t1(core->getBaseType(thm.getLHS()));
  Type t2(core->getBaseType(thm.getRHS()));
  DebugAssert(t1 == t2, "TheoryCore::rewriteCore: type-unsafe rewrite:\n"
	      "type1 = "+t1.toString()+"\ntype2 = "+t2.toString()+"\n\n"
	      +thm.toString());
}
#endif

bool TheoryCore::incomplete(vector<string>& reasons) {
  if(d_incomplete.size() > 0) {
    for(CDMap<string,bool>::iterator i=d_incomplete.begin(),
	  iend=d_incomplete.end(); i!=iend; ++i)
      reasons.push_back((*i).first);
    return true;
  }
  else
    return false;
}


Theorem TheoryCore::rewriteCore(const Expr& e)
{
  TRACE("rewrite", "rewriteCore(", e, ") {");
  if (e.hasFind()) {
    TRACE("rewrite", "rewriteCore[hasFind] => ", find(e), " }");
    IF_DEBUG(checkRewriteType(this, find(e)));
    return find(e);
  }
  if (e.isRewriteNormal()) {
    IF_DEBUG(
      // Check that the RewriteNormal flag is set properly.  Note that we
      // assume theory-specific rewrites are idempotent
      e.clearRewriteNormal();
      Expr rewritten(rewriteCore(e).getRHS());
      e.setRewriteNormal(); // Restore the flag
      DebugAssert(rewritten == e, 
		  "Expected no change: e = " + e.toString()
		  +"\n rewriteCore(e) = "+rewritten.toString());
    )
    TRACE("rewrite", "rewriteCore[isRewriteNormal] => ", reflexivityRule(e), " }");
    return reflexivityRule(e);
  }
  Theorem thm;
  if (e.isNot() || e.isEq()) {
    TRACE("rewrite", "rewriteLitCore(", e, ") {");
    thm = rewriteLitCore(e);
    IF_DEBUG(checkRewriteType(this, thm));
    TRACE("rewrite", "rewriteLitCore => ", thm, "}");
    if(e != thm.getRHS()) {
      thm = rewriteCore(thm);
      TRACE("rewrite", "rewriteCore => ", thm, " }");
      return thm;
    }
  }
  TRACE("rewrite", ("rewrite["+theoryOf(e)->getName()+"]("), e, ") {");
  thm = theoryOf(e)->rewrite(e);
  TRACE("rewrite", ("rewrite["+theoryOf(e)->getName()+"] => "), thm, "}");
  IF_DEBUG(checkRewriteType(this, thm));

  const Expr& e2 = thm.getRHS();

  // Theory-specific rewrites for equality should ensure that lhs >= rhs, or
  // there is danger of an infinite loop.
  DebugAssert(!e2.isEq() || e2[0] >= e2[1],
	      "theory-specific rewrites for equality should ensure lhs >= rhs");

  if (e != e2) {
    thm = rewriteCore(thm);
    TRACE("rewrite", "rewriteCore => ", thm, " }");
    return thm;
  }

  TRACE("rewrite", "rewriteCore[refl] => ", thm, " }");
  return thm;
}


Theorem TheoryCore::rewriteCore(const Theorem& e) {
  DebugAssert(e.isRewrite(),
	      "rewriteCore(thm): not equality or iff:\n  " + e.toString());
  return transitivityRule(e, rewriteCore(e.getRHS()));
}


Theorem TheoryCore::rewriteLiteral(const Expr& e) {
  DebugAssert(isLiteral(e), "rewriteLiteral("+e.toString()
	      +")\nExpected a literal.");
  Theory* i = theoryOf(e);
  Theorem res;
  bool neg(e.isNot());
  const Expr& a = neg? e[0] : e;
  TRACE("rewrite", ("rewriteAtomic["+i->getName()+"]("), a, ") {");
  res = i->rewriteAtomic(a);
  TRACE("rewrite", ("rewriteAtomic["+i->getName()+"] => "), res, "}");
  if(neg) res = iffContrapositive(res);
  return res;
}

/*****************************************************************************/
/*!
 * Function: TheoryCore::solve
 *
 * Author: Clark Barrett
 *
 * Created: Wed Feb 26 16:17:54 2003
 *
 * This is a generalization of what's in my thesis.  The goal is to rewrite e
 * into an equisatisfiable conjunction of equations such that the left-hand
 * side of each equation is a variable which does not appear as an i-leaf of
 * the rhs, where i is the theory of the primary solver.  Any solution which
 * satisfies this is fine.  "Solvers" from other theories can do whatever they
 * want as long as we eventually reach this form.
 */
/*****************************************************************************/
Theorem TheoryCore::solve(const Theorem& eThm)
{
  const Expr& e = eThm.getExpr();
  TRACE("facts solve", "solve[Core](", e, ") {");
  Theorem thm;
  Expr e2;
  if (d_solver) {
    TRACE("facts solve", "solve["+d_solver->getName()+"](", e,") {");
    thm = d_solver->solve(eThm);
    TRACE("facts solve", "solve["+d_solver->getName()+"] => ", thm.getExpr(), " }");
    e2 = thm.getExpr();
    // FIXME: this stops other solvers from being run when arithmetic
    // solves into a system of equations.  This results in
    // INCOMPLETENESS.  In my opinion (Sergey), the correct fix is to
    // allow multiple independent solvers, just like rewriters.
    if (!e2.isEq() || theoryOf(e2) == d_solver) {
      TRACE("facts solve", "solve[Core](non-eq) => ", thm, " }");
      // setIncomplete("Not all solvers have been run (this is a bug)");
      return thm;
    }
  }
  else {
    thm = eThm;
    e2 = e;
  }
  Theory* i = theoryOf(e2);
  TRACE("facts solve", "solve["+i->getName()+"](", e2,") {");
  Theorem thm2 = i->solve(thm);
  TRACE("facts solve", "solve["+i->getName()+"] => ", thm2.getExpr(), " }");
  IF_DEBUG(
    // Check that the result is in solved form according to the primary solver
    if (d_solver) {
      e2 = thm2.getExpr();
      if(e2.isEq()) {
	Expr solved(d_solver->solve(thm2).getExpr());
	DebugAssert(solved == e2, "e2 = "+e2.toString()
		    +"\nsolved = "+solved.toString());
      }
      // FIXME: how do we check it for AND and EXISTS?
    }
    );
  TRACE("facts solve", "solve[Core] => ", thm2, " }");
  return thm2;
}

Theorem TheoryCore::subtypePredicate(const Expr& e) {
  TRACE("tccs", "subtypePredicate(", e, ") {");
  Theorem pred(e.lookupSubtypePred());
  if(pred.isNull()) {
    vector<Theorem> thms;
    thms.push_back(typePred(e));
    TRACE("tccs", "subtypePredicate: typePred(e) = ", thms[0].getExpr(), "");
    for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i)
      thms.push_back(subtypePredicate(*i));
    pred = andIntro(thms);
    pred = rewriteAnd(pred);
    e.setSubtypePred(pred);
  }
  TRACE("tccs", "subtypePredicate => ", pred.getExpr(), " }");
  return pred;
}


Theorem TheoryCore::preprocess(const Expr& e)
{
  Theorem thm1 = pushNegation(e);
  thm1 = transitivityRule(thm1, simplifyThm(thm1.getRHS(), true));
  if(d_vcl->getFlags()["ite"].getBool())
    thm1 = transitivityRule(thm1,ite_convert(thm1.getRHS()));
  if(d_vcl->getFlags()["pp"].getBool())
    return transitivityRule(thm1,ite_simplify(thm1.getRHS()));
  return thm1;
}


Theorem TheoryCore::preprocess(const Theorem& thm)
{
  return iffMP(thm, preprocess(thm.getExpr()));
}


void TheoryCore::setFindLiteral(const Theorem& thm, bool notifySAT) {
  const Expr& e = thm.getExpr();
  if (e.isNot()) {
    if (!e[0].hasFind()) {
      if(notifySAT) enqueueSE(thm);
      e[0].setFind(d_commonRules->notToIff(thm));
      d_em->invalidateSimpCache();
    }
  }
  else if (!e.hasFind()) {
    e.setFind(d_commonRules->iffTrue(thm));
    if(notifySAT) enqueueSE(thm);
    d_em->invalidateSimpCache();
  }
}


// Recursively descend into the expression e, keeping track of whether
// we are under even or odd number of negations ('neg' == true means
// odd, the context is "negative").

// Produce a proof of e <==> e' or !e <==> e', depending on whether
// neg is false or true, respectively.
Theorem TheoryCore::pushNegationRec(const Expr& e, bool neg) {
  TRACE("pushNegation", "pushNegationRec(", e,
	", neg=" + string(neg? "true":"false") + ") {");
  DebugAssert(!isTerm(e), "pushNegationRec: not boolean e = "+e.toString());
  ExprMap<Theorem>::iterator i = d_pushNegCache.find(neg? !e : e);
  if(i != d_pushNegCache.end()) { // Found cached result
    TRACE("pushNegation", "pushNegationRec [cached] => ", (*i).second, "}");
    return (*i).second;
  }
  // By default, do not rewrite
  Theorem res(reflexivityRule((neg)? !e : e));
  if(neg) {
    switch(e.getKind()) {
      case TRUE: res = d_commonRules->rewriteNotTrue(!e); break;
      case FALSE: res = d_commonRules->rewriteNotFalse(!e); break;
      case NOT: 
        res = pushNegationRec(d_commonRules->rewriteNotNot(!e), false);
        break;
      case AND:
        res = pushNegationRec(d_commonRules->rewriteNotAnd(!e), false);
        break;
      case OR: 
        res = pushNegationRec(d_commonRules->rewriteNotOr(!e), false);
        break;
      case IMPLIES: {
        vector<Theorem> thms;
        thms.push_back(d_commonRules->rewriteImplies(e));
        res = pushNegationRec
          (substitutivityRule(getOp(!e), thms), true);
        break;
      }
      case IFF:
	// Preserve equivalences to explore structural similarities
	if(e[0].getKind() == e[1].getKind())
	  res = reflexivityRule(!e);
	else
	  res = pushNegationRec(d_commonRules->rewriteNotIff(!e), false);
        break;
      case ITE:
        res = pushNegationRec(d_commonRules->rewriteNotIte(!e), false);
        break;

      // Replace LETDECL and CONSTDEF with their definitions.  The
      // typechecker makes sure it's type-safe to do so.
      case LETDECL: {
        vector<Theorem> thms;
        thms.push_back(d_commonRules->rewriteLetDecl(e));
        res = pushNegationRec
          (substitutivityRule(getOp(!e), thms), true);
        break;
      }
      case CONSTDEF:
      {
        vector<Theorem> thms;
        thms.push_back(d_commonRules->rewriteConstDef(e));
        res = pushNegationRec
          (substitutivityRule(getOp(!e), thms), true);
        break;
      }
      default:
        res = reflexivityRule(!e);
    } // end of switch(e.getKind())
  } else { // if(!neg)
    switch(e.getKind()) {
      case NOT: res = pushNegationRec(e[0], true); break;
      case AND:
      case OR:
      case IFF:
      case ITE: {
        Op op = getOp(e);
        vector<Theorem> thms;
        for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i)
          thms.push_back(pushNegationRec(*i, false));
        res = substitutivityRule(op, thms);
        break;
      }
      case IMPLIES:
        res = pushNegationRec(d_commonRules->rewriteImplies(e), false);
        break;
      case LETDECL:
        res = pushNegationRec(d_commonRules->rewriteLetDecl(e), false);
        break;
      case CONSTDEF:
        res = pushNegationRec(d_commonRules->rewriteConstDef(e), false);
        break;
      default:
        res = reflexivityRule(e);
    } // end of switch(e.getKind())
  }
  TRACE("pushNegation", "pushNegationRec => ", res, "}");
  d_pushNegCache[neg? !e : e] = res;
  return res;
}


Theorem TheoryCore::pushNegationRec(const Theorem& thm, bool neg) {
  DebugAssert(thm.isRewrite(), "pushNegationRec(Theorem): bad theorem: "
	      + thm.toString());
  Expr e(thm.getRHS());
  if(neg) {
    DebugAssert(e.isNot(), 
		"pushNegationRec(Theorem, neg = true): bad Theorem: "
		+ thm.toString());
    e = e[0];
  }
  return transitivityRule(thm, pushNegationRec(e, neg));
}


Theorem TheoryCore::rewriteIte(const Expr& e) {
  if (e[0].isTrue())
    return d_commonRules->rewriteIteTrue(e);
  if (e[0].isFalse())
    return d_commonRules->rewriteIteFalse(e);
  if (e[1] == e[2])
    return d_commonRules->rewriteIteSame(e);
  // This results in exponential blowup
//   if(!isTerm(e1) && !isTerm(e2))
//     return d_commonRules->rewriteIteBool(c, e1, e2);
  // Default case: no rewrite
  return reflexivityRule(e);
}


// We assume that the cache is initially empty
Theorem TheoryCore::pushNegation(const Expr& e) {
  if(isTerm(e)) return reflexivityRule(e);
  Theorem res(pushNegationRec(e, false));
  d_pushNegCache.clear();
  return res;
}


Theorem TheoryCore::skolemize(const Theorem& thm) {
  return d_vcl->getSE()->skolemize(thm, true);
}


Theorem TheoryCore::varIntroSkolem(const Expr& e) {
  return d_vcl->getSE()->varIntroSkolem(e, true);
}


bool
TheoryCore::isPropAtom(const Expr& e) {
  switch(e.getKind()) {
  case AND:
  case OR:
  case IFF:
  case IMPLIES:
  case ITE:
  case NOT:
    return false;
  default:
    return true;
  }
}

bool
TheoryCore::isPropLiteral(const Expr& e) {
  switch(e.getKind()) {
  case NOT:
    return isPropAtom(e[0]);
  default:
    return isPropAtom(e);
  }
}


void TheoryCore::assignValue(const Expr& t, const Expr& val) {
  DebugAssert(d_varAssignments.count(t) == 0
	      || d_varAssignments[t].getRHS() == val,
	      "TheoryCore::assignValue("+t.toString()+" := "+val.toString()
	      +")\n variable already has a different value");
  // Check if the assignment theorem can be derived from the find of t
  Theorem thm(find(t));
  const Expr& t2 = thm.getRHS();

  if(t2!=val) {
    bool isBool(getType(t2).isBool());
    Expr assertion = (isBool)? t2.iffExpr(val) : t2.eqExpr(val);
    Theorem assertThm(d_vcl->getSE()->newAssertion(assertion));
//     if(isBool || findExpr(val)!=val)
      addFact(assertThm);
//     else addEquality(assertThm);
    thm = transitivityRule(thm, assertThm);
  }
  d_varAssignments[t] = thm;
}


void TheoryCore::assignValue(const Theorem& thm) {
  DebugAssert(thm.isRewrite(), "TheoryCore::assignValue("+thm.toString()+")");
  const Expr& t = thm.getLHS();
  const Expr& val = thm.getRHS();
  DebugAssert(d_varAssignments.count(t) == 0
	      || d_varAssignments[t].getRHS() == val,
	      "TheoryCore::assignValue("+thm.toString()
	      +")\n variable already has a different value");
  d_varAssignments[t] = thm;
  // Check if the assignment theorem can be derived from the find of t
  Theorem findThm(find(t));
  const Expr& t2 = findThm.getRHS();

  if(t2!=val) {
    Theorem thm2 = transitivityRule(symmetryRule(findThm), thm);
//     bool isBool(getType(t2).isBool());
//     if(isBool || findExpr(val)!=val)
      addFact(thm2);
//     else addEquality(thm2);
  }
}


bool TheoryCore::getResource() {
  // Unlimited resource
  if(d_resourceLimit == 0) return true;
  // Resource is limited: check and consume it
  if(d_resourceLimit > 1) {
    d_resourceLimit--;
    return true;
  }
  else
    return false;
}
