/*****************************************************************************/
/*!
 *\file theory_datatype.cpp
 *
 * Author: Clark Barrett
 *
 * Created: Wed Dec  1 22:32:26 2004
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#include "theory_datatype.h"
#include "datatype_proof_rules.h"
#include "vcl.h"
#include "typecheck_exception.h"
#include "parser_exception.h"
#include "smtlib_exception.h"


using namespace std;
using namespace CVCL;


///////////////////////////////////////////////////////////////////////////////
// TheoryDatatype Public Methods                                             //
///////////////////////////////////////////////////////////////////////////////


TheoryDatatype::TheoryDatatype(VCL *vcl)
  : Theory(vcl, "Datatypes"), d_vcl(vcl)
{
  d_rules = createProofRules(vcl, this);

  // Register new local kinds with ExprManager
  getEM()->newKind(DATATYPE, "DATATYPE", true);
  getEM()->newKind(THISTYPE, "THISTYPE");
  getEM()->newKind(CONSTRUCTOR, "CONSTRUCTOR");
  getEM()->newKind(SELECTOR, "SELECTOR");
  getEM()->newKind(TESTER, "TESTER");

  vector<int> kinds;
  kinds.push_back(DATATYPE);
  kinds.push_back(THISTYPE);
  kinds.push_back(TESTER);
  kinds.push_back(CONSTRUCTOR);
  kinds.push_back(SELECTOR);

  registerTheory(this, kinds);

  d_thisType = Type(newExpr(getEM(), THISTYPE));
}


TheoryDatatype::~TheoryDatatype() {
  delete d_rules;
}


void TheoryDatatype::assertFact(const Theorem& e)
{
}


void TheoryDatatype::checkSat(bool fullEffort)
{
}


Theorem TheoryDatatype::rewrite(const Expr& e)
{
  return reflexivityRule(e);
}


void TheoryDatatype::setup(const Expr& e)
{
}


void TheoryDatatype::update(const Theorem& e, const Expr& d)
{
}


void TheoryDatatype::computeType(const Expr& e)
{
}


void TheoryDatatype::computeModelTerm(const Expr& e, std::vector<Expr>& v) {
}



Expr TheoryDatatype::computeTCC(const Expr& e)
{
  return trueExpr();
}

///////////////////////////////////////////////////////////////////////////////
// Pretty-printing			                                     //
///////////////////////////////////////////////////////////////////////////////


ExprStream& TheoryDatatype::print(ExprStream& os, const Expr& e) {
  return os;
}

//////////////////////////////////////////////////////////////////////////////
//parseExprOp:
//translating special Exprs to regular EXPR??
///////////////////////////////////////////////////////////////////////////////
Expr TheoryDatatype::parseExprOp(const Expr& e)
{
  // If the expression is not a list, it must have been already
  // parsed, so just return it as is.
  if(RAW_LIST != e.getKind()) return e;

  DebugAssert(e.arity() > 0,
	      "TheoryDatatype::parseExprOp:\n e = "+e.toString());
  
  DebugAssert(e[0].getKind() == ID,
              "Expected ID kind for first elem in list expr");

  const Expr& c1 = e[0][0];
  int kind = getEM()->getKind(c1.getString());
  switch(kind) {
    case DATATYPE: {
      DebugAssert(e.arity() > 1,
                  "Empty DATATYPE expression\n"
                  " (expected at least one constructor): "+e.toString());
      vector<string> constructorNames;
      vector<vector<string> > selectorNames(e.arity()-1);
      vector<vector<Type> > types(e.arity()-1);
      Expr constructor, selectors, selector;
      int i,j;

      for(i = 1; i < e.arity(); ++i) {
        constructor = e[i];
        DebugAssert(constructor.arity() == 1 || constructor.arity() == 2,
                    "Expected arity 1 or 2 for constructor"
                    +constructor.toString());
        DebugAssert(constructor[0].getKind() == ID,
                    "Expected ID kind for constructor name"
                    +constructor.toString());
        constructorNames.push_back(constructor[0][0].getString());
        if (constructor.arity() == 2) {
          selectors = constructor[1];
          DebugAssert(selectors.getKind() == RAW_LIST,
                      "List expected as second argument of constructor:\n"
                      +constructor.toString());
          for (j = 0; j < selectors.arity(); ++j) {
            selector = selectors[j];
            DebugAssert(selector.arity() == 2,
                        "Expected arity 2 for selector"
                        +selector.toString());
            DebugAssert(selector[0].getKind() == ID,
                        "Expected ID kind for selector name"
                        +selector.toString());
            selectorNames[i-1].push_back(selector[0][0].getString());
            if (selector[1].getKind() == ID &&
                selector[1][0].getString() == "THISTYPE") {
              types[i-1].push_back(d_thisType);
            }
            else types[i-1].push_back(Type(parseExpr(selector[1])));
          }
        }
      }

      Type dt = dataType(constructorNames, selectorNames, types);
      return dt.getExpr();
    }

    case THISTYPE: {
      throw ParserException("Use of keyword THISTYPE outside of DATATYPE declaration");
      break;
    }

    default: {
      throw ParserException("Unexpected datatype expression: "+e.toString());
    }
  }
  return e;
}


Type TheoryDatatype::dataType(const vector<string>& constructors,
                              const vector<vector<string> >& selectors,
                              const vector<vector<Type> >& types)
{
  vector<Expr> constructorExprs, selectorExprs;
  vector<Expr> datatypeList, constructorList;
  bool wellFounded = false, infinite = false, thisWellFounded;

  if (constructors.size() != selectors.size() ||
      selectors.size() != types.size()) {
    throw TypecheckException
      ("dataType: must have same number of constructors and selector and type vectors");
  }


  for (size_t i = 0, iend=constructors.size(); i < iend; ++i) {
    if (selectors[i].size() != types[i].size()) {
      throw TypecheckException
        ("dataType: constructor "+constructors[i]+
         ": number of selectors does not match number of types");
    }
    thisWellFounded = true;
    selectorExprs.clear();
    const vector<string>& sels = selectors[i];
    const vector<Type>& tps = types[i];
    for (size_t j = 0, jend=sels.size(); j < jend; ++j) {
      Expr selExpr = newNamedExpr(SELECTOR, sels[j], tps[j].getExpr());
      selectorExprs.push_back(selExpr);
      if (tps[j] == d_thisType) {
        infinite = true;
        thisWellFounded = false;
      }
    }
    if (thisWellFounded) wellFounded = true;
    Expr constrExpr=newNamedExpr(CONSTRUCTOR, constructors[i], selectorExprs);
    constructorExprs.push_back(constrExpr);
  }

  if (!wellFounded) {
    throw TypecheckException("Error: data type is not well-founded");
  }

  if (!infinite) {
    throw TypecheckException("Error: finite data types are not yet supported");
  }

  // construct datatype
  Type res = Type(newExpr(getEM(), DATATYPE, constructorExprs));

  /*
  // construct operators for constructors and testers
  vector<Type> constructorArgs;
  Expr tester;
  for (i = 0; i < constructorExprs.size(); ++i) {
    constructorArgs.clear();
    for (j = 0; j < types[i].size(); ++j) {
      if (types[i][j].isNull()) constructorArgs.push_back(res);
      else constructorArgs.push_back(types[i][j]);
    }
    constructorExprs[i].setType(funType(constructorArgs, res));
    // register as global symbol

    tester = newExpr(TESTER, newExpr(getEM(), STRING_EXPR, constructors[i]));
    tester.setType(funType(res.funType(boolType())));
    // register as global symbol
  }

  // construct type for selectors
  Type dom;
  for (i = 0; i < selectors.size(); ++i) {
    for (j = 0; j < selectors[i].size(); ++j) {
      selectorExprs[k].setType(
        res.funType(types[i][j].isNull() ? res : types[i][j]));
      // register as global symbol
      k++
    }
  }
  */

  return res;
}


Expr TheoryDatatype::datatypeConsExpr(const string& constructor,
                                      const vector<Expr>& args)
{
  return Expr();
}


Expr TheoryDatatype::datatypeSelExpr(const string& selector, const Expr& arg)
{
  return Expr();
}
